# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Sayan Patra
"""Functions to run benchmarking."""

import itertools
import os
import timeit
from pathlib import Path

from greykite.common.evaluation import EvaluationMetricEnum
from greykite.framework.templates.autogen.forecast_config import EvaluationPeriodParam
from greykite.framework.templates.autogen.forecast_config import ForecastConfig
from greykite.framework.templates.autogen.forecast_config import ModelComponentsParam
from greykite.framework.templates.forecaster import Forecaster
from greykite.framework.templates.model_templates import ModelTemplateEnum


def benchmark_silverkite_template(
        data_name,
        df,
        forecast_horizons,
        fit_algorithms,
        max_cvs,
        metadata=None,
        evaluation_metric=None):
    """Benchmarks silverkite template and returns the output as a list

    :param data_name: str
        Name of the dataset we are performing benchmarking on
        For real datasets, the data_name matches the corresponding filename in the data/ folder
        For simulated datasets, we follow the convention "<freq>_simulated" e.g. "daily_simulated"
    :param df: pd.DataFrame
        Dataframe containing the time and value columns
    :param forecast_horizons: List[int]
        One forecast is created for every given forecast_horizon
    :param fit_algorithms: List[str]
        Names of predictive models to fit.
        Options are "linear", "lasso", "ridge", "rf" etc.
    :param max_cvs: List[int] or None
        Number of maximum CV folds to use.
    :param metadata: :class:`~greykite.framework.templates.autogen.forecast_config.MetadataParam` or None, default None
        Information about the input data. See
        :class:`~greykite.framework.templates.autogen.forecast_config.MetadataParam`.
    :param evaluation_metric: :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam` or None, default None
        What metrics to evaluate. See
        :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam`.
    :return: .csv file
        Each row of the .csv file records the following outputs from one run of the silverkite template:

            - "data_name": Fixed string "<freq>_simulated", or name of the dataset in data/ folder
            - "forecast_model_name": "silverkite_<fit_algorithm>" e.g. "silverkite_linear" or "prophet"
            - "train_period": train_period
            - "forecast_horizon": forecast_horizon
            - "fit_algorithm": fit algorithm name
            - "cv_folds": max_cv
            - "runtime_sec": runtime in seconds
            - "train_mae": Mean Absolute Error of training data in backtest
            - "train_mape": Mean Absolute Percent Error of training data in backtest
            - "test_mae": Mean Absolute Error of testing data in backtest
            - "test_mape": Mean Absolute Percent Error of testing data in backtest
    """
    benchmark_results = []

    for forecast_horizon, fit_algorithm, max_cv in itertools.product(forecast_horizons, fit_algorithms, max_cvs):
        model_components = ModelComponentsParam(
            custom={
                "fit_algorithm_dict": {
                    "fit_algorithm": fit_algorithm,
                },
                "feature_sets_enabled": True
            }
        )
        evaluation_period = EvaluationPeriodParam(
            cv_max_splits=max_cv
        )

        start_time = timeit.default_timer()
        forecaster = Forecaster()
        result = forecaster.run_forecast_config(
            df=df,
            config=ForecastConfig(
                model_template=ModelTemplateEnum.SILVERKITE.name,
                forecast_horizon=forecast_horizon,
                metadata_param=metadata,
                evaluation_metric_param=evaluation_metric,
                model_components_param=model_components,
                evaluation_period_param=evaluation_period,
            )
        )
        runtime = timeit.default_timer() - start_time

        output_dict = dict(
            data_name=data_name,
            forecast_model_name=f"silverkite_{fit_algorithm}",
            train_period=df.shape[0],
            forecast_horizon=forecast_horizon,
            cv_folds=result.grid_search.n_splits_,
            runtime_sec=round(runtime, 3),
            train_mae=result.backtest.train_evaluation["MAE"].round(3),
            train_mape=result.backtest.train_evaluation["MAPE"].round(3),
            test_mae=result.backtest.test_evaluation["MAE"].round(3),
            test_mape=result.backtest.test_evaluation["MAPE"].round(3)
        )
        benchmark_results.append(output_dict)

    return benchmark_results


def get_default_benchmark_real_datasets():
    """Default parameter sets to framework.benchmark real datasets. The datasets are located in data folder.
    Every tuple has the following structure:
    (data_name, frequency, time_col, value_col, forecast_horizon)"""
    real_datasets = [
        # daily_peyton_manning, 8 years of data
        ("daily_peyton_manning", "D", "ts", "y", [30, 365]),
        # daily_female_births, 1 year of data
        ("daily_female_births", "D", "Date", "Births", [30, 3*30])
    ]

    return real_datasets


def get_default_benchmark_silverkite_parameters():
    """Default parameter sets for benchmarking silverkite template"""
    return dict(
        fit_algorithms=("linear", "lasso", "ridge", "rf", "sgd"),
        max_cvs=[3]
    )


def get_default_benchmark_parameters():
    """Default parameter sets for benchmarking"""
    directory = Path(__file__).parents[4]  # src/ root
    directory = os.path.abspath(directory)
    data_directory = os.path.join(directory, "benchmark_output")
    return dict(
        metric=EvaluationMetricEnum.MeanSquaredError,
        data_directory=data_directory
    )


def get_default_benchmark_simulated_datasets():
    """Default parameter sets to generate simulated data for benchmarking.
    The training periods and forecast horizon are chosen to complement default real datasets.
    Every tuple has the following structure:
    (data_name, frequency, training_periods, forecast_horizon)"""
    simulation_parameters = [
        # daily data
        ("daily_simulated", "D", 3*30, [30]),
        ("daily_simulated", "D", 2*365, [365]),
        # hourly data
        ("hourly_simulated", "H", 7*24, [24]),
        ("hourly_simulated", "H", 30*24, [7*24]),
        ("hourly_simulated", "H", 365*24, [6*30*24]),
        ("hourly_simulated", "H", 4*365*24, [365*24])
    ]

    return simulation_parameters



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen, Rachit Kumar, Sayan Patra
"""sklearn estimator for fbprophet"""

import fbprophet
from fbprophet.plot import plot_components
from sklearn.exceptions import NotFittedError
from sklearn.metrics import mean_squared_error

from greykite.common.constants import PREDICTED_COL
from greykite.common.constants import PREDICTED_LOWER_COL
from greykite.common.constants import PREDICTED_UPPER_COL
from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.logging import pprint
from greykite.sklearn.estimator.base_forecast_estimator import BaseForecastEstimator


class ProphetEstimator(BaseForecastEstimator):
    """Wrapper for Facebook Prophet model.

    Parameters
    ----------
    score_func : callable
        see BaseForecastEstimator

    coverage : float between [0.0, 1.0]
        see BaseForecastEstimator

    null_model_params : dict with arguments to define DummyRegressor null model, optional, default=None
        see BaseForecastEstimator

    add_regressor_dict: dictionary of extra regressors to be added to the model, optional, default=None
        These should be available for training and entire prediction interval.

        Dictionary format::

            add_regressor_dict={  # we can add as many regressors as we want, in the following format
                "reg_col1": {
                    "prior_scale": 10,
                    "standardize": True,
                    "mode": 'additive'
                },
                "reg_col2": {
                    "prior_scale": 20,
                    "standardize": True,
                    "mode": 'multiplicative'
                }
            }

    add_seasonality_dict: dict of custom seasonality parameters to be added to the model, optional, default=None
        parameter details: https://github.com/facebook/prophet/blob/master/python/fbprophet/forecaster.py - refer to
        add_seasonality() function.
        Key is the seasonality component name e.g. 'monthly'; parameters are specified via dict.

        Dictionary format::

            add_seasonality_dict={
                'monthly': {
                    'period': 30.5,
                    'fourier_order': 5
                },
                'weekly': {
                    'period': 7,
                    'fourier_order': 20,
                    'prior_scale': 0.6,
                    'mode': 'additive',
                    'condition_name': 'condition_col'  # takes a bool column in df with True/False values. This means that
                    # the seasonality will only be applied to dates where the condition_name column is True.
                },
                'yearly': {
                    'period': 365.25,
                    'fourier_order': 10,
                    'prior_scale': 0.2,
                    'mode': 'additive'
                }
            }

        Note: If there is a conflict in built-in and custom seasonality e.g. both have "yearly", then custom seasonality
        will be used and Model will throw a warning such as:
        "INFO:fbprophet:Found custom seasonality named "yearly", disabling built-in yearly seasonality."

    kwargs : additional parameters
        Other parameters are the same as Prophet model, with one exception:
        ``interval_width`` is specified by ``coverage``.

        See source code ``__init__`` for the parameter names, and refer to
        Prophet documentation for a description:

            * https://facebook.github.io/prophet/docs/quick_start.html
            * https://github.com/facebook/prophet/blob/c7a150312ad26db5a8ea81d8cd32dd5fd6ce1eda/python/fbprophet/forecaster.py#L30

    Attributes
    ----------
    model : ``Prophet`` object
        Prophet model object
    forecast : `pandas.DataFrame`
        Output of predict method of ``Prophet``.
    """

    def __init__(
            self,
            score_func=mean_squared_error,
            coverage=0.80,  # to specify interval_width in Prophet
            null_model_params=None,
            growth="linear",
            changepoints=None,
            n_changepoints=25,
            changepoint_range=0.8,
            yearly_seasonality="auto",
            weekly_seasonality="auto",
            daily_seasonality="auto",
            holidays=None,
            seasonality_mode="additive",
            seasonality_prior_scale=10.0,
            holidays_prior_scale=10.0,
            changepoint_prior_scale=0.05,
            mcmc_samples=0,
            uncertainty_samples=1000,
            add_regressor_dict=None,
            add_seasonality_dict=None):
        # every subclass of BaseForecastEstimator must call super().__init__
        super().__init__(
            score_func=score_func,
            coverage=coverage,
            null_model_params=null_model_params)

        # necessary to set parameters, to ensure get_params() works (used in grid search)
        self.score_func = score_func
        self.coverage = coverage
        self.null_model_params = null_model_params
        self.growth = growth
        self.changepoints = changepoints
        self.n_changepoints = n_changepoints
        self.changepoint_range = changepoint_range
        self.yearly_seasonality = yearly_seasonality
        self.weekly_seasonality = weekly_seasonality
        self.daily_seasonality = daily_seasonality
        self.holidays = holidays
        self.seasonality_mode = seasonality_mode
        self.seasonality_prior_scale = seasonality_prior_scale
        self.holidays_prior_scale = holidays_prior_scale
        self.changepoint_prior_scale = changepoint_prior_scale
        self.mcmc_samples = mcmc_samples
        self.uncertainty_samples = uncertainty_samples
        # additional regressor names and optimization
        self.add_regressor_dict = add_regressor_dict
        # additional seasonality parameters
        self.add_seasonality_dict = add_seasonality_dict
        # set by the fit method
        self.model = None
        # set by the predict method
        self.forecast = None

    def fit(self, X, y=None, time_col=TIME_COL, value_col=VALUE_COL, **fit_params):
        """Fits fbprophet model.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input timeseries, with timestamp column,
            value column, and any additional regressors.
            The value column is the response, included in
            X to allow transformation by `sklearn.pipeline.Pipeline`
        y : ignored
            The original timeseries values, ignored.
            (The y for fitting is included in ``X``.)
        time_col : `str`
            Time column name in ``X``
        value_col : `str`
            Value column name in ``X``
        fit_params : `dict`
            additional parameters for null model

        Returns
        -------
        self : self
            Fitted model is stored in ``self.model``.
        """
        super().fit(X, y=y, time_col=time_col, value_col=value_col, **fit_params)

        if self.add_regressor_dict is None:
            fit_columns = [time_col, value_col]
        else:
            reg_cols = list(self.add_regressor_dict.keys())
            fit_columns = [time_col, value_col] + reg_cols

        fit_df = X.reset_index(drop=True)[fit_columns]
        fit_df.rename(columns={time_col: "ds", value_col: "y"}, inplace=True)
        # Prophet expects these column names. Other estimators can use TIME_COL, etc.
        # uses coverage instead of interval_width to set prediction band width. This ensures a common
        # interface for parameters common to every BaseForecastEstimator, usually also needed for forecast evaluation
        # model must be initialized here, not in __init__, to update parameters in grid search
        self.model = fbprophet.Prophet(
            growth=self.growth,
            changepoints=self.changepoints,
            n_changepoints=self.n_changepoints,
            changepoint_range=self.changepoint_range,
            yearly_seasonality=self.yearly_seasonality,
            weekly_seasonality=self.weekly_seasonality,
            daily_seasonality=self.daily_seasonality,
            holidays=self.holidays,
            seasonality_mode=self.seasonality_mode,
            seasonality_prior_scale=self.seasonality_prior_scale,
            holidays_prior_scale=self.holidays_prior_scale,
            changepoint_prior_scale=self.changepoint_prior_scale,
            mcmc_samples=self.mcmc_samples,
            interval_width=self.coverage,
            uncertainty_samples=self.uncertainty_samples
        )
        # if extra regressors are given, we add them to temporal features data
        # This implementation assumes that the regressor(s) are provided in time series df, alongside target column.
        if self.add_regressor_dict is not None:
            for reg_col, reg_params in self.add_regressor_dict.items():
                self.model.add_regressor(name=reg_col, **reg_params)

        # if custom seasonality is provided, we supply it to Prophet model
        if self.add_seasonality_dict is not None:
            for seasonality_type, seasonality_params in self.add_seasonality_dict.items():
                self.model.add_seasonality(name=seasonality_type, **seasonality_params)

        self.model.fit(fit_df)
        return self

    def predict(self, X, y=None):
        """Creates forecast for dates specified in ``X``.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input timeseries with timestamp column and any additional regressors.
            Timestamps are the dates for prediction.
            Value column, if provided in X, is ignored.
        y : ignored

        Returns
        -------
        predictions : `pandas.DataFrame`
            Forecasted values for the dates in ``X``. Columns:

                * TIME_COL dates
                * PREDICTED_COL predictions
                * PREDICTED_LOWER_COL lower bound of predictions, optional
                * PREDICTED_UPPER_COL upper bound of predictions, optional
                * [other columns], optional

            PREDICTED_LOWER_COL and PREDICTED_UPPER_COL are present iff coverage is not None
        """
        # Returns the cached result if applicable
        cached_predictions = super().predict(X=X)
        if cached_predictions is not None:
            return cached_predictions

        # if regressors are not provided, then use time column to predict future. Else, use regressor from predict df
        if self.add_regressor_dict is None:
            predict_columns = [self.time_col_]
        else:
            reg_cols = list(self.add_regressor_dict.keys())
            predict_columns = [self.time_col_] + reg_cols

        fut_df = X.reset_index(drop=True)[predict_columns]
        # prophet expects time_col name to be "ds"
        fut_df.rename(columns={self.time_col_: "ds"}, inplace=True)

        pred_df = self.model.predict(fut_df)
        self.forecast = pred_df     # This is used by the plot_components

        # renames columns to standardized schema
        output_columns = {
            "ds": TIME_COL,
            "yhat": PREDICTED_COL,
            "yhat_lower": PREDICTED_LOWER_COL,
            "yhat_upper": PREDICTED_UPPER_COL
        }

        predictions = (pred_df[output_columns.keys()]
                       .rename(output_columns, axis=1))
        # Caches the predictions
        self.cached_predictions_ = predictions
        return predictions

    def summary(self):
        """Prints input parameters and Prophet model parameters.

        Returns
        -------
        log_message : str
            log message printed to logging.info()
        """
        super().summary()
        if self.model is not None:
            log_message(pprint(self.model.params), LoggingLevelEnum.INFO)

    def plot_components(
            self,
            uncertainty=True,
            plot_cap=True,
            weekly_start=0,
            yearly_start=0,
            figsize=None):
        """Plot the ``Prophet`` forecast components on the dataset passed
        to ``predict``.

        Will plot whichever are available of: trend, holidays, weekly
        seasonality, and yearly seasonality.

        Parameters
        ----------
        uncertainty : `bool`, optional, default True
            Boolean to plot uncertainty intervals.
        plot_cap : `bool`, optional, default True
            Boolean indicating if the capacity should be shown
            in the figure, if available.
        weekly_start : `int`, optional, default 0
            Specifying the start day of the weekly seasonality plot.
            0 (default) starts the week on Sunday.
            1 shifts by 1 day to Jan 2, and so on.
        yearly_start : `int`, optional, default 0
            Specifying the start day of the yearly seasonality plot.
            0 (default) starts the year on Jan 1.
            1 shifts by 1 day to Jan 2, and so on.
        figsize : `tuple` , optional, default None
            Width, height in inches.

        Returns
        -------
        fig: `matplotlib.figure.Figure`
            A matplotlib figure.
        """
        if self.model is None:
            raise NotFittedError("The fit method has not been run yet.")

        if self.forecast is None:
            raise RuntimeError("The predict method has not been run yet.")

        try:
            return plot_components(
                m=self.model,
                fcst=self.forecast,
                uncertainty=uncertainty,
                plot_cap=plot_cap,
                weekly_start=weekly_start,
                yearly_start=yearly_start,
                figsize=figsize)
        except AttributeError as e:
            if "'DatetimeIndex'" in repr(e):
                # 'DatetimeIndex' object has no attribute 'weekday_name'
                raise Exception("Prophet 0.5 component plots are incompatible with pandas 1.*. "
                                "Upgrade to fbprophet:0.6 or higher.")
            else:
                raise e



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini, Albert Chen
"""Evaluation functions.
Valid input processing is done within these evaluation functions
so they can be called from anywhere without error checks.
"""

import math
import warnings
from collections import namedtuple
from enum import Enum
from functools import partial

import numpy as np
import pandas as pd
from scipy.stats import pearsonr
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import median_absolute_error
from sklearn.metrics import r2_score

from greykite.common.constants import ACTUAL_COL
from greykite.common.constants import COVERAGE_VS_INTENDED_DIFF
from greykite.common.constants import LOWER_BAND_COVERAGE
from greykite.common.constants import PREDICTED_COL
from greykite.common.constants import PREDICTED_LOWER_COL
from greykite.common.constants import PREDICTED_UPPER_COL
from greykite.common.constants import PREDICTION_BAND_COVERAGE
from greykite.common.constants import PREDICTION_BAND_WIDTH
from greykite.common.constants import UPPER_BAND_COVERAGE
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message


def all_equal_length(*arrays):
    """Checks whether input arrays have equal length.

    :param arrays: one or more lists/numpy arrays/pd.Series
    :return: true if lengths are all equal, otherwise false
    """
    length = None
    for arr in arrays:
        try:
            if length is not None and len(arr) != length:
                return False
            length = len(arr)
        except TypeError:  # continue if `arr` has no len method. constants and None are excluded from length check
            pass
    return True


def valid_elements_for_evaluation(reference_array, *arrays):
    """Keeps finite elements from reference_array, and corresponding elements in *arrays.

    :param reference_array: np.array, pd.Series, or list
    :param arrays: additional array-like objects of the same length
        can be None or numeric, in which case original value is returned
    :return: list of numpy arrays with valid indices [reference_array, *arrays]
    """
    if not all_equal_length(reference_array, *arrays):
        raise Exception("length of arrays do not match")

    reference_array = np.array(reference_array)
    # Finds the indices of finite elements in reference array
    keep = np.isfinite(reference_array)
    num_remaining = keep.sum()
    num_removed = reference_array.shape[0] - num_remaining
    if num_remaining == 0:
        warnings.warn(f"There are 0 non-null elements for evaluation.")
    if num_removed > 0:
        warnings.warn(f"{num_removed} value(s) in y_true were NA or infinite and are omitted in error calc.")

    # Keeps these indices in all arrays. Leaves float, int, and None as-is
    return [reference_array[keep]] + [np.array(array)[keep] if (array is not None
                                                                and not isinstance(array, float)
                                                                and not isinstance(array, int))
                                      else array
                                      for array in arrays]


def aggregate_array(ts_values, agg_periods=7, agg_func=np.sum):
    """Aggregates input array.
    Divides array from left to right into bins of size agg_periods, and applies agg_func to each block.
    Drops records from the left if needed to ensure all bins are full.

    :param ts_values: list, np.array, or pd.Series to aggregate
    :param agg_periods: number of periods to combine in aggregation
    :param agg_func: aggregation function, e.g. np.max, np.sum. Must take an array and returns a number
    :return: array, aggregated so that every agg_periods periods are combined into one

    Examples:
    >>> aggregate_array([1.0, 2.0, 3.0, 4.0], 2, np.sum)
    array([3., 7.])
    >>> aggregate_array(pd.Series([1.0, 2.0, 3.0, 4.0, 5.0]), 2, np.sum)
    array([5., 9.])
    >>> aggregate_array(np.array([1.0, 2.0, 3.0, 4.0, 5.0]), 2, np.max)
    array([3., 5.])
    """
    ts_values = np.array(ts_values)
    n_periods = len(ts_values)

    drop_first_periods = n_periods % agg_periods  # drop these periods from the front, to ensure all bins are full
    if drop_first_periods == n_periods:
        drop_first_periods = 0
        warnings.warn(f"Requested agg_periods={agg_periods}, but there are only {n_periods}. Using all for aggregation")
    elif drop_first_periods > 0:
        log_message(f"Requested agg_periods={agg_periods} for data of length {n_periods}. Dropping first"
                    f" {drop_first_periods} records before aggregation", LoggingLevelEnum.INFO)

    # creates dummy time index for aggregation
    dates = pd.date_range("2018-01-01", periods=n_periods - drop_first_periods, freq="1D")
    ts = pd.Series(ts_values[drop_first_periods:], index=dates)
    aggregated_array = ts.resample(f"{agg_periods}D", closed="left") \
        .agg(lambda x: agg_func(x)) \
        .values
    return aggregated_array


def add_preaggregation_to_scorer(score_func, agg_periods=7, agg_func=np.sum):
    """Takes a scorer and returns a scorer that pre-aggregates input.

    :param score_func: function that maps two arrays to a number. E.g. (y_true, y_pred) -> error
    :param agg_periods: number of periods to combine in aggregation
    :param agg_func: function that takes an array and returns a number, e.g. np.max, np.sum. This function should
        handle np.nan without error
    :return: scorer that applies agg_func to data before calling score_func

    For example, if have daily data and want to evaluate MSE on weekly totals, the appropriate scorer is:
        add_preaggregation_to_scorer(mean_squared_error, agg_periods=7, agg_func=np.max)
    """
    def score_func_preagg(y_true, y_pred, **kwargs):
        y_true_agg = aggregate_array(y_true, agg_periods, agg_func)
        y_pred_agg = aggregate_array(y_pred, agg_periods, agg_func)
        return score_func(y_true_agg, y_pred_agg, **kwargs)
    return score_func_preagg


def add_finite_filter_to_scorer(score_func):
    """Takes a scorer and returns a scorer that ignores NA / infinite elements in y_true.

    sklearn scorers (and others) don't handle arrays with 0 length. In that case, return None

    :param score_func: function that maps two arrays to a number. E.g. (y_true, y_pred) -> error
    :return: scorer that drops records where y_true is not finite
    """
    def score_func_finite(y_true, y_pred, **kwargs):
        y_true, y_pred = valid_elements_for_evaluation(y_true, y_pred)
        if len(y_true) == 0:  # returns None if there are no elements
            return None
        return score_func(y_true, y_pred, **kwargs)
    return score_func_finite


def r2_null_model_score(
        y_true,
        y_pred,
        y_pred_null=None,
        y_train=None,
        loss_func=mean_squared_error):
    """Calculates improvement in the loss function compared
    to the predictions of a null model. Can be used to evaluate
    model quality with respect to a simple baseline model.

    The score is defined as::

        R2_null_model_score = 1.0 - loss_func(y_true, y_pred) / loss_func(y_true, y_pred_null)

    Parameters
    ----------
    y_true : `list` [`float`] or `numpy.array`
        Observed response (usually on a test set).
    y_pred : `list` [`float`] or `numpy.array`
        Model predictions (usually on a test set).
    y_pred_null : `list` [`float`] or `numpy.array` or None
        A baseline prediction model to compare against.
        If None, derived from ``y_train`` or ``y_true``.
    y_train : `list` [`float`] or `numpy.array` or None
        Response values in the training data.
        If ``y_pred_null`` is None, then ``y_pred_null`` is set to the mean of ``y_train``.
        If ``y_train`` is also None, then ``y_pred_null`` is set to the mean of ``y_true``.
    loss_func : callable, default `sklearn.metrics.mean_squared_error`
        The error loss function with signature (true_values, predicted_values).

    Returns
    -------
    r2_null_model : `float`
        A value within (-\\infty, 1.0]. Higher scores are better.
        Can be interpreted as the improvement in the loss function
        compared to the predictions of the null model.
        For example, a score of 0.74 means the loss is 74% lower
        than for the null model.

    Notes
    -----
    There is a connection between ``R2_null_model_score`` and ``R2``.
    ``R2_null_model_score`` can be interpreted as the additional improvement in the
    coefficient of determination (i.e. ``R2``, see `sklearn.metrics.r2_score`) with
    respect to a null model.

    Under the default settings of this function, where ``loss_func`` is
    mean squared error and ``y_pred_null`` is the average of
    ``y_true``, the scores are equivalent::

        # simplified definition of R2_score, where SSE is sum of squared error
        y_true_avg = np.repeat(np.average(y_true), y_true.shape[0])
        R2_score := 1.0 - SSE(y_true, y_pred) / SSE(y_true, y_true_avg)
        R2_score := 1.0 - MSE(y_true, y_pred) / VAR(y_true)  # equivalent definition

        r2_null_model_score(y_true, y_pred) == r2_score(y_true, y_pred)

    ``r2_score`` is 0 if simply predicting the mean (y_pred = y_true_avg).

    If ``y_pred_null`` is passed, and if ``loss_func`` is mean squared error
    and ``y_true`` has nonzero variance, this function measures how much "r2_score of the
    predictions (``y_pred``)" closes the gap between "r2_score of the null model (``y_pred``)"
    and the "r2_score of the best possible model (``y_true``)", which is 1.0::

        R2_pred = r2_score(y_true, y_pred)       # R2 of predictions
        R2_null = r2_score(y_pred_null, y_pred)  # R2 of null model
        r2_null_model_score(y_true, y_pred, y_pred_null) == (R2_pred - R2_null) / (1.0 - R2_null)

    When ``y_pred_null=y_true_avg``, ``R2_null`` is 0 and this reduces to the formula above.

    Summary (for ``loss_func=mean_squared_error``):

        - If ``R2_null>0`` (good null model), then ``R2_null_model_score < R2_score``
        - If ``R2_null=0`` (uninformative null model), then ``R2_null_model_score = R2_score``
        - If ``R2_null<0`` (poor null model), then ``R2_null_model_score > R2_score``

    For other loss functions, ``r2_null_model_score`` has the same connection to pseudo R2.
    """
    y_true, y_pred, y_train, y_pred_null = valid_elements_for_evaluation(y_true, y_pred, y_train, y_pred_null)
    r2_null_model = None

    if len(y_true) > 0:
        # determines null model
        if y_pred_null is None and y_train is not None:  # from training data
            y_pred_null = np.repeat(y_train.mean(), len(y_true))
        elif y_pred_null is None:  # from test data
            y_pred_null = np.repeat(y_true.mean(), len(y_true))
        elif isinstance(y_pred_null, float) or isinstance(y_pred_null, int):  # constant null model
            y_pred_null = np.repeat(y_pred_null, len(y_true))
        # otherwise, y_pred_null is an array, used directly

        model_loss = loss_func(y_true, y_pred)
        null_loss = loss_func(y_true, y_pred_null)

        if null_loss > 0.0:
            r2_null_model = 1.0 - (model_loss / null_loss)
    return r2_null_model


def calc_pred_err(y_true, y_pred):
    """Calculates the basic error measures in
    `~greykite.common.evaluation.EvaluationMetricEnum`
    and returns them in a dictionary.

    Parameters
    ----------
    y_true : `list` [`float`] or `numpy.array`
        Observed values.
    y_pred : `list` [`float`] or `numpy.array`
        Model predictions.

    Returns
    -------
    error : `dict` [`str`, `float` or None]
        Dictionary mapping
        `~greykite.common.evaluation.EvaluationMetricEnum`
        metric names to their values for the given ``y_true``
        and ``y_pred``.

        The dictionary is empty is there are no finite elements
        in ``y_true``.
    """
    y_true, y_pred = valid_elements_for_evaluation(y_true, y_pred)
    error = {}

    if len(y_true) > 0:
        for enum in EvaluationMetricEnum:
            metric_name = enum.get_metric_name()
            metric_func = enum.get_metric_func()
            error.update({metric_name: metric_func(y_true, y_pred)})
    return error


@add_finite_filter_to_scorer
def mean_absolute_percent_error(y_true, y_pred):
    """Calculates mean absolute percentage error.

    :param y_true: observed values given in a list (or numpy array)
    :param y_pred: predicted values given in a list (or numpy array)
    :return: mean absolute percent error
    """
    smallest_denominator = np.min(np.abs(y_true))
    if smallest_denominator == 0:
        warnings.warn("y_true contains 0. MAPE is undefined.")
        return None
    elif smallest_denominator < 1e-8:
        warnings.warn("y_true contains very small values. MAPE is likely highly volatile.")
    return 100 * (np.abs(y_true - y_pred) / np.abs(y_true)).mean()


@add_finite_filter_to_scorer
def median_absolute_percent_error(y_true, y_pred):
    """Calculates median absolute percentage error.

    :param y_true: observed values given in a list (or numpy array)
    :param y_pred: predicted values given in a list (or numpy array)
    :return: median absolute percent error
    """
    smallest_denominator = np.min(np.abs(y_true))
    if smallest_denominator == 0:
        warnings.warn("y_true contains 0. MedAPE is undefined.")
        return None
    elif smallest_denominator < 1e-8:
        num_small = len(y_true[y_true < 1e-8])
        if num_small > (len(y_true) - 1) // 2:
            # if too many very small values, median is affected
            warnings.warn("y_true contains very small values. MedAPE is likely highly volatile.")
    return 100 * np.median((np.abs(y_true - y_pred) / np.abs(y_true)))


@add_finite_filter_to_scorer
def symmetric_mean_absolute_percent_error(y_true, y_pred):
    """Calculates symmetric mean absolute percentage error
    Note that we do not include a factor of 2 in the denominator, so the range is 0% to 100%.

    :param y_true: observed values given in a list (or numpy array)
    :param y_pred: predicted values given in a list (or numpy array)
    :return: symmetric mean absolute percent error
    """
    smallest_denominator = np.min(np.abs(y_true) + np.abs(y_pred))
    if smallest_denominator == 0:
        warnings.warn("denominator contains 0. sMAPE is undefined.")
        return None
    elif smallest_denominator < 1e-8:
        warnings.warn("denominator contains very small values. sMAPE is likely highly volatile.")
    return 100 * (np.abs(y_true - y_pred) / (np.abs(y_true) + np.abs(y_pred))).mean()


@add_finite_filter_to_scorer
def root_mean_squared_error(y_true, y_pred):
    """Calculates root mean square error.

    :param y_true: observed values given in a list (or numpy array)
    :param y_pred: predicted values given in a list (or numpy array)
    :return: mean absolute percent error
    """
    return math.sqrt(mean_squared_error(y_true, y_pred))


@add_finite_filter_to_scorer
def correlation(y_true, y_pred):
    """Calculates correlation.

    :param y_true: observed values given in a list (or numpy array)
    :param y_pred: predicted values given in a list (or numpy array)
    :return: correlation
    """
    if np.unique(y_true).size == 1:
        warnings.warn("y_true is constant. Correlation is not defined.")
        return None
    elif np.unique(y_pred).size == 1:
        warnings.warn("y_pred is constant. Correlation is not defined.")
        return None
    else:
        return pearsonr(y_true, y_pred)[0]


@add_finite_filter_to_scorer
def quantile_loss(y_true, y_pred, q: float = 0.95):
    """Calculates the quantile (pinball) loss with quantile q of y_true and y_pred.

    :param y_true: one column of true values.
    :type y_true: list or numpy array
    :param y_pred: one column of predicted values.
    :type y_pred: list or numpy array.
    :param q: the quantile.
    :type q: float
    :return: quantile loss.
    :returns: float
    """
    return np.where(y_true < y_pred, (1 - q) * (y_pred - y_true), q * (y_true - y_pred)).mean()


def quantile_loss_q(q):
    """Returns quantile loss function for the specified quantile"""
    def quantile_loss_wrapper(y_true, y_pred):
        return quantile_loss(y_true, y_pred, q=q)
    return quantile_loss_wrapper


def fraction_within_bands(observed, lower, upper):
    """Computes the fraction of observed values between lower and upper.

    :param observed: pd.Series or np.array, numeric, observed values
    :param lower: pd.Series or np.array, numeric, lower bound
    :param upper: pd.Series or np.array, numeric, upper bound
    :return: float between 0.0 and 1.0
    """
    observed, lower, upper = valid_elements_for_evaluation(observed, lower, upper)
    num_reversed = np.count_nonzero(upper < lower)
    if num_reversed > 0:
        warnings.warn(f"{num_reversed} of {len(observed)} upper bound values are smaller than the lower bound")
    return np.count_nonzero((observed > lower) & (observed < upper)) / len(observed) if len(observed) > 0 else None


@add_finite_filter_to_scorer
def fraction_outside_tolerance(y_true, y_pred, rtol=0.05):
    """Returns the fraction of predicted values whose relative difference
    from the true value is strictly greater than ``rtol``.

    Parameters
    ----------
    y_true : `list` or `numpy.array`
        True values.
    y_pred : `list` or `numpy.array`
        Predicted values.
    rtol : `float`, default 0.05
        Relative error tolerance.
        For example, 0.05 allows for 5% relative error.

    Returns
    -------
    frac_outside_tolerance : float
        Fraction of values outside tolerance. (0.0 to 1.0)
        A value is outside tolerance if `numpy.isclose`
        with the specified ``rtol`` and ``atol=0.0`` returns False.
    """
    return np.mean(~np.isclose(y_pred, y_true, rtol=rtol, atol=0.0))


def prediction_band_width(observed, lower, upper):
    """Computes the prediction band width, expressed as a % relative to observed.
    Width is defined as average ratio of (upper-lower)/observed.

    :param observed: pd.Series or np.array, numeric, observed values
    :param lower: pd.Series or np.array, numeric, lower bound
    :param upper: pd.Series or np.array, numeric, upper bound
    :return: float, average percentage width
    """
    observed, lower, upper = valid_elements_for_evaluation(observed, lower, upper)
    observed = np.abs(observed)
    num_reversed = np.count_nonzero(upper < lower)
    if num_reversed > 0:
        warnings.warn(f"{num_reversed} of {len(observed)} upper bound values are smaller than the lower bound")
    # if there are 0s in observed, relative size is undefined
    return 100.0 * np.mean(np.abs(upper - lower) / observed) if len(observed) > 0 and observed.min() > 0 else None


def calc_pred_coverage(observed, predicted, lower, upper, coverage):
    """Calculates the prediction coverages:
   prediction band width, prediction band coverage etc.

    :param observed: pd.Series or np.array, numeric, observed values
    :param predicted: pd.Series or np.array, numeric, predicted values
    :param lower: pd.Series or np.array, numeric, lower bound
    :param upper: pd.Series or np.array, numeric, upper bound
    :param coverage: float, intended coverage of the prediction bands (0.0 to 1.0)
    :return: prediction band width, prediction band coverage etc.
    """
    observed, predicted, lower, upper = valid_elements_for_evaluation(
        observed,
        predicted,
        lower,
        upper)
    metrics = {}

    if len(observed) > 0:
        # relative size of prediction bands vs actual, as a percent
        enum = ValidationMetricEnum.BAND_WIDTH
        metric_func = enum.get_metric_func()
        metrics.update({PREDICTION_BAND_WIDTH: metric_func(observed, lower, upper)})

        enum = ValidationMetricEnum.BAND_COVERAGE
        metric_func = enum.get_metric_func()
        # fraction of observations within the bands
        metrics.update({PREDICTION_BAND_COVERAGE: metric_func(observed, lower, upper)})
        # fraction of observations within the lower band
        metrics.update({LOWER_BAND_COVERAGE: metric_func(observed, lower, predicted)})
        # fraction of observations within the upper band
        metrics.update({UPPER_BAND_COVERAGE: metric_func(observed, predicted, upper)})
        # difference between actual and intended coverage
        metrics.update({COVERAGE_VS_INTENDED_DIFF: (metrics[PREDICTION_BAND_COVERAGE] - coverage)})
    return metrics


def elementwise_residual(true_val, pred_val):
    """The residual between a single true and predicted value.

    Parameters
    ----------
    true_val : float
        True value.
    pred_val : float
        Predicted value.

    Returns
    -------
    residual : float
        The residual, true minus predicted
    """
    return true_val - pred_val


def elementwise_absolute_error(true_val, pred_val):
    """The absolute error between a single true and predicted value.

    Parameters
    ----------
    true_val : float
        True value.
    pred_val : float
        Predicted value.

    Returns
    -------
    residual : float
        Absolute error, |true_val - pred_val|
    """
    return abs(true_val - pred_val)


def elementwise_squared_error(true_val, pred_val):
    """The absolute error between a single true and predicted value.

    Parameters
    ----------
    true_val : float
        True value.
    pred_val : float
        Predicted value.

    Returns
    -------
    residual : float
        Squared error, (true_val - pred_val)^2
    """
    return (true_val - pred_val)**2


def elementwise_absolute_percent_error(true_val, pred_val):
    """The absolute percent error between a single true and predicted value.

    Parameters
    ----------
    true_val : float
        True value.
    pred_val : float
        Predicted value.

    Returns
    -------
    percent_error : float
        Percent error, pred_val / true_val - 1
    """
    if true_val == 0:
        warnings.warn("true_val is 0. Percent error is undefined.")
        return np.nan
    elif true_val < 1e-8:
        warnings.warn("true_val is less than 1e-8. Percent error is very likely highly volatile.")
    return 100 * abs(true_val - pred_val) / abs(true_val)


def elementwise_quantile(true_val, pred_val, q):
    """The quantile loss between a single true and predicted value.

    Parameters
    ----------
    true_val : float
        True value.
    pred_val : float
        Predicted value.

    Returns
    -------
    quantile_loss : float
        Quantile loss, absolute error weighed by ``q``
        for underpredictions and ``1-q`` for overpredictions.
    """
    weight = 1 - q if true_val < pred_val else q
    return weight * abs(true_val - pred_val)


def elementwise_outside_tolerance(true_val, pred_val, rtol=0.05):
    """Whether the relative difference between ``pred_val`` and
    ``true_val`` is strictly greater than ``rtol``.

    Parameters
    ----------
    true_val : float
        True value.
    pred_val : float
        Predicted value.
    rtol : float, default 0.05
        Relative error tolerance.
        For example, 0.05 allows for 5% relative error.

    Returns
    -------
    is_outside_tolerance : float
        1.0 if the error is outside tolerance, else 0.0.
        A value is outside tolerance if `numpy.isclose`
        with the specified ``rtol`` and ``atol=0.0`` returns False.

    See Also
    --------

    """
    return 0.0 if np.isclose(pred_val, true_val, rtol=rtol, atol=0.0) else 1.0


def elementwise_within_bands(true_val, lower_val, upper_val):
    """Whether ``true_val`` is strictly between ``lower_val`` and ``upper_val``.

    Parameters
    ----------
    true_val : float
        True value.
    lower_val : float
        Lower bound.
    upper_val : float
        Upper bound.

    Returns
    -------
    is_within_band : float
        1.0 if error is strictly within the limits, else 0.0
    """
    return 1.0 if lower_val < true_val < upper_val else 0.0


class ElementwiseEvaluationMetricEnum(Enum):
    """Evaluation metrics for a single element.
    This is the function computed on each row, before aggregation across records.
    The aggregation function can be mean, median, quantile, max, sqrt of the mean, etc.

    For example, AbsoluteError followed by mean aggregation gives MeanAbsoluteError.
    EvaluationMetricEnum is more efficient if it can be used directly.
    """
    ElementwiseEvaluationMetric = namedtuple("ElementwiseEvaluationMetric", "func, name, args")
    """
    The elementwise evaluation function and associated metadata.

        ``"func"`` : `callable`
            Elementwise function.
        ``"name"`` : `str`
            Short name for the metric.
        ``"args"`` : `list` [`str`]
            Description of function arguments.
            e.g. [actual, predicted]
    """
    Residual = ElementwiseEvaluationMetric(
        elementwise_residual,
        "residual",
        [ACTUAL_COL, PREDICTED_COL])
    """Residual, true-pred"""
    AbsoluteError = ElementwiseEvaluationMetric(
        elementwise_absolute_error,
        "absolute_error",
        [ACTUAL_COL, PREDICTED_COL])
    """Absolute error, abs(true-pred)"""
    SquaredError = ElementwiseEvaluationMetric(
        elementwise_squared_error,
        "squared_error",
        [ACTUAL_COL, PREDICTED_COL])
    """Squared error, (true-pred)^2"""
    AbsolutePercentError = ElementwiseEvaluationMetric(
        elementwise_absolute_percent_error,
        "absolute_percent_error",
        [ACTUAL_COL, PREDICTED_COL])
    """Percent error, abs(true-pred)/abs(true)"""
    Quantile80 = ElementwiseEvaluationMetric(
        partial(elementwise_quantile, q=0.80),
        "quantile_loss_80",
        [ACTUAL_COL, PREDICTED_COL])
    """Quantile loss with q=0.80"""
    Quantile90 = ElementwiseEvaluationMetric(
        partial(elementwise_quantile, q=0.90),
        "quantile_loss_90",
        [ACTUAL_COL, PREDICTED_COL])
    """Quantile loss with q=0.90"""
    Quantile95 = ElementwiseEvaluationMetric(
        partial(elementwise_quantile, q=0.95),
        "quantile_loss_95",
        [ACTUAL_COL, PREDICTED_COL])
    """Quantile loss with q=0.95"""
    Quantile99 = ElementwiseEvaluationMetric(
        partial(elementwise_quantile, q=0.99),
        "quantile_loss_99",
        [ACTUAL_COL, PREDICTED_COL])
    """Quantile loss with q=0.99"""
    OutsideTolerance5 = ElementwiseEvaluationMetric(
        partial(elementwise_outside_tolerance, rtol=0.05),
        "outside_tolerance_5p",
        [ACTUAL_COL, PREDICTED_COL])
    """Whether the predicted value deviates more than 5% from the true value"""
    Coverage = ElementwiseEvaluationMetric(
        elementwise_within_bands,
        "coverage",
        [ACTUAL_COL, PREDICTED_LOWER_COL, PREDICTED_UPPER_COL])
    """Whether the actual value is within the prediction interval"""

    def get_metric_func(self):
        """Returns the metric function"""
        return self.value.func

    def get_metric_name(self):
        """Returns the short name"""
        return self.value.name

    def get_metric_args(self):
        """Returns the expected argument list"""
        return self.value.args


class EvaluationMetricEnum(Enum):
    """Valid evaluation metrics.
    The values tuple is ``(score_func: callable, greater_is_better: boolean, short_name: str)``

    ``add_finite_filter_to_scorer`` is added to the metrics that are directly imported from
    ``sklearn.metrics`` (e.g. ``mean_squared_error``) to ensure that the metric gets calculated
    even when inputs have missing values.
    """
    Correlation = (correlation, True, "CORR")
    """Pearson correlation coefficient between forecast and actuals. Higher is better."""
    CoefficientOfDetermination = (add_finite_filter_to_scorer(r2_score), True, "R2")
    """Coefficient of determination. See `sklearn.metrics.r2_score`. Higher is better. Equals `1.0 - mean_squared_error / variance(actuals)`."""
    MeanSquaredError = (add_finite_filter_to_scorer(mean_squared_error), False, "MSE")
    """Mean squared error, the average of squared differences,
    see `sklearn.metrics.mean_squared_error`."""
    RootMeanSquaredError = (root_mean_squared_error, False, "RMSE")
    """Root mean squared error, the square root of `sklearn.metrics.mean_squared_error`"""
    MeanAbsoluteError = (add_finite_filter_to_scorer(mean_absolute_error), False, "MAE")
    """Mean absolute error, average of absolute differences,
    see `sklearn.metrics.mean_absolute_error`."""
    MedianAbsoluteError = (add_finite_filter_to_scorer(median_absolute_error), False, "MedAE")
    """Median absolute error, median of absolute differences,
    see `sklearn.metrics.median_absolute_error`."""
    MeanAbsolutePercentError = (mean_absolute_percent_error, False, "MAPE")
    """Mean absolute percent error, error relative to actuals expressed as a %,
    see `wikipedia MAPE <https://en.wikipedia.org/wiki/Mean_absolute_percentage_error>`_."""
    MedianAbsolutePercentError = (median_absolute_percent_error, False, "MedAPE")
    """Median absolute percent error, median of error relative to actuals expressed as a %,
    a median version of the MeanAbsolutePercentError, less affected by extreme values."""
    SymmetricMeanAbsolutePercentError = (symmetric_mean_absolute_percent_error, False, "sMAPE")
    """Symmetric mean absolute percent error, error relative to (actuals+forecast) expressed as a %.
    Note that we do not include a factor of 2 in the denominator, so the range is 0% to 100%,
    see `wikipedia sMAPE <https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error>`_."""
    Quantile80 = (quantile_loss_q(0.80), False, "Q80")
    """Quantile loss with q=0.80::

        np.where(y_true < y_pred, (1 - q) * (y_pred - y_true), q * (y_true - y_pred)).mean()
    """
    Quantile95 = (quantile_loss_q(0.95), False, "Q95")
    """Quantile loss with q=0.95::

        np.where(y_true < y_pred, (1 - q) * (y_pred - y_true), q * (y_true - y_pred)).mean()
    """
    Quantile99 = (quantile_loss_q(0.99), False, "Q99")
    """Quantile loss with q=0.99::

        np.where(y_true < y_pred, (1 - q) * (y_pred - y_true), q * (y_true - y_pred)).mean()
    """
    FractionOutsideTolerance1 = (partial(fraction_outside_tolerance, rtol=0.01), False, "OutsideTolerance1p")
    """Fraction of forecasted values that deviate more than 1% from the actual"""
    FractionOutsideTolerance2 = (partial(fraction_outside_tolerance, rtol=0.02), False, "OutsideTolerance2p")
    """Fraction of forecasted values that deviate more than 2% from the actual"""
    FractionOutsideTolerance3 = (partial(fraction_outside_tolerance, rtol=0.03), False, "OutsideTolerance3p")
    """Fraction of forecasted values that deviate more than 3% from the actual"""
    FractionOutsideTolerance4 = (partial(fraction_outside_tolerance, rtol=0.04), False, "OutsideTolerance4p")
    """Fraction of forecasted values that deviate more than 4% from the actual"""
    FractionOutsideTolerance5 = (partial(fraction_outside_tolerance, rtol=0.05), False, "OutsideTolerance5p")
    """Fraction of forecasted values that deviate more than 5% from the actual"""

    def get_metric_func(self):
        """Returns the metric function"""
        return self.value[0]

    def get_metric_greater_is_better(self):
        """Returns the greater_is_better boolean"""
        return self.value[1]

    def get_metric_name(self):
        """Returns the short name"""
        return self.value[2]


class ValidationMetricEnum(Enum):
    """Valid diagnostic metrics.
    The values tuple is ``(score_func: callable, greater_is_better: boolean)``
    """
    BAND_WIDTH = (prediction_band_width, False)
    BAND_COVERAGE = (fraction_within_bands, True)

    def get_metric_func(self):
        return self.value[0]

    def get_metric_greater_is_better(self):
        return self.value[1]



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Sayan Patra
"""Color palette for plotting."""

from plotly.colors import DEFAULT_PLOTLY_COLORS
from plotly.colors import n_colors
from plotly.colors import validate_colors


def get_color_palette(num, colors=DEFAULT_PLOTLY_COLORS):
    """Returns ``num`` of distinct RGB colors.
    If ``num`` is less than or equal to the length of ``colors``, first ``num``
    elements of ``colors`` are returned.
    Else ``num`` elements of colors are interpolated between the first and the last
    colors of ``colors``.

    Parameters
    ----------
    num : `int`
        Number of colors required.
    colors : [`str`, `list` [`str`]], default ``DEFAULT_PLOTLY_COLORS``
        Which colors to use to build the color palette.
        This can be a list of RGB colors or a `str` from ``PLOTLY_SCALES``.

    Returns
    -------
    color_palette: List
        A list consisting ``num`` of RGB colors.
    """
    validate_colors(colors, colortype="rgb")
    if len(colors) == 1:
        return colors * num
    elif len(colors) >= num:
        color_palette = colors[0:num]
    else:
        color_palette = n_colors(
            colors[0],
            colors[-1],
            num,
            colortype="rgb")
    return color_palette



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

from typing import List
from typing import Union

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class DtypeColumnSelector(BaseEstimator, TransformerMixin):
    """Simple selector that subsets a DataFrame to the columns of interest by their type
    This allows a Pipeline that applies different transformations to subsets of columns,
       and unions that result
    See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.select_dtypes.html

    :param include: A selection of dtypes or strings to be included. At least
        one of include/exclude parameters must be supplied.
    :type include: scalar or list-like
    :param exclude: A selection of dtypes or strings to be excluded. At least
        one of include/exclude parameters must be supplied.
    :type exclude: scalar or list-like
    """
    def __init__(self, include: Union[str, List[str]] = None, exclude: Union[str, List[str]] = None):
        self.include = include
        self.exclude = exclude

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        assert isinstance(X, pd.DataFrame)
        return X.select_dtypes(include=self.include, exclude=self.exclude)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen


from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import pandas as pd

from greykite.algo.changepoint.adalasso.changepoint_detector import get_changepoints_dict
from greykite.algo.forecast.silverkite.constants.silverkite_column import SilverkiteColumn
from greykite.algo.forecast.silverkite.constants.silverkite_constant import SilverkiteConstant
from greykite.algo.forecast.silverkite.constants.silverkite_constant import default_silverkite_constant
from greykite.algo.forecast.silverkite.constants.silverkite_holiday import SilverkiteHoliday
from greykite.algo.forecast.silverkite.constants.silverkite_time_frequency import SilverkiteTimeFrequencyEnum
from greykite.algo.forecast.silverkite.forecast_silverkite import SilverkiteForecast
from greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper import cols_interact
from greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper import generate_holiday_events
from greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper import get_event_pred_cols
from greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper import patsy_categorical_term
from greykite.common import constants as cst
from greykite.common.constants import GROWTH_COL_ALIAS
from greykite.common.enums import SimpleTimeFrequencyEnum
from greykite.common.enums import TimeEnum
from greykite.common.features.timeseries_features import get_available_holidays_across_countries
from greykite.common.features.timeseries_features import get_changepoint_features_and_values_from_config
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import unique_elements_in_list
from greykite.common.python_utils import update_dictionary
from greykite.common.time_properties_forecast import get_forecast_time_properties


class SimpleSilverkiteForecast(SilverkiteForecast):
    """A derived class of `~greykite.algo.forecast.silverkite.SilverkiteForecast`.
    Provides an alternative interface with simplified configuration parameters.
    Produces the same trained model output and uses the same predict functions.
    """
    def __init__(
            self,
            constants: SilverkiteConstant = default_silverkite_constant):
        super().__init__(constants=constants)
        self._silverkite_time_frequency_enum: Type[SilverkiteTimeFrequencyEnum] = constants.get_silverkite_time_frequency_enum()
        self._silverkite_holiday: Type[SilverkiteHoliday] = constants.get_silverkite_holiday()
        self._silverkite_column: Type[SilverkiteColumn] = constants.get_silverkite_column()

    def convert_params(
            self,
            df: pd.DataFrame,
            time_col: str,
            value_col: str,
            time_properties: Optional[Dict] = None,
            freq: Optional[str] = None,
            forecast_horizon: Optional[int] = None,
            origin_for_time_vars: Optional[float] = None,
            train_test_thresh: Optional[datetime] = None,
            training_fraction: Optional[float] = 0.9,
            fit_algorithm: str = "ridge",
            fit_algorithm_params: Optional[Dict] = None,
            holidays_to_model_separately: Optional[Union[str, List[str]]] = "auto",
            holiday_lookup_countries: Optional[Union[str, List[str]]] = "auto",
            holiday_pre_num_days: int = 2,
            holiday_post_num_days: int = 2,
            holiday_pre_post_num_dict: Optional[Dict] = None,
            daily_event_df_dict: Optional[Dict] = None,
            changepoints_dict: Optional[Dict] = None,
            yearly_seasonality: Union[bool, str, int] = "auto",
            quarterly_seasonality: Union[bool, str, int] = "auto",
            monthly_seasonality: Union[bool, str, int] = "auto",
            weekly_seasonality: Union[bool, str, int] = "auto",
            daily_seasonality: Union[bool, str, int] = "auto",
            max_daily_seas_interaction_order: Optional[int] = None,
            max_weekly_seas_interaction_order: Optional[int] = None,
            autoreg_dict: Optional[Dict] = None,
            seasonality_changepoints_dict: Optional[Dict] = None,
            min_admissible_value: Optional[float] = None,
            max_admissible_value: Optional[float] = None,
            uncertainty_dict: Optional[Dict] = None,
            growth_term: Optional[str] = "linear",
            regressor_cols: Optional[List[str]] = None,
            feature_sets_enabled: Optional[Union[bool, str, Dict[str, Optional[Union[bool, str]]]]] = "auto",
            extra_pred_cols: Optional[List[str]] = None,
            regression_weight_col: Optional[str] = None,
            simulation_based: Optional[bool] = False):
        """Converts parameters of
        :func:`~greykite.algo.forecast.silverkite.forecast_simple_silverkite` into those
        of :func:`~greykite.algo.forecast.forecast_silverkite.SilverkiteForecast::forecast`.

        Makes it easier to set parameters to ``SilverkiteForecast::forecast`` suitable for most forecasting problems.
        Provides data-aware defaults for seasonality and interaction terms. Provides a simple
        configuration of holidays from an internal holiday database, and user-friendly configuration
        for growth and regressors.

        These parameters can be set from a plain-text config (e.g. no pandas dataframes).
        The parameter list is intentionally flat to facilitate hyperparameter grid search. Every
        parameter is either a parameter of ``SilverkiteForecast::forecast`` or a tuning parameter.

        Notes
        -----
        The basic parameters are identical to ``SilverkiteForecast::forecast``.
        The more complex parameters are specified via config parameters:

        * ``daily_event_df_dict`` (via ``holiday*``)
        * ``fs_components_df`` (via `*_seasonality``)
        * ``extra_pred_cols`` (via ``holiday*``, ``*seas*``, ``growth_term``,
          ``regressor_cols``, ``feature_sets_enabled``, ``extra_pred_cols``)

        Parameters
        ----------
        df : `pandas.DataFrame`
            A data frame which includes the timestamp column
            as well as the value column. This is the ``df`` for
            training the model, not for future prediction.
        time_col : `str`
            The column name in `df` representing time for the time series data
            The time column can be anything that can be parsed by pandas DatetimeIndex
        value_col: `str`
            The column name which has the value of interest to be forecasted
        time_properties : `dict` [`str`, `any`] or None, optional
            Time properties dictionary (likely produced by
            `~greykite.common.time_properties_forecast.get_forecast_time_properties`)
            with keys:

                ``"ts"`` : `UnivariateTimeSeries` or None
                    ``df`` converted to a ``UnivariateTimeSeries``.
                ``"period"`` : `int`
                    Period of each observation (i.e. minimum time between observations, in seconds).
                ``"simple_freq"`` : `SimpleTimeFrequencyEnum`
                    ``SimpleTimeFrequencyEnum`` member corresponding to data frequency.
                ``"num_training_points"`` : `int`
                    Number of observations for training.
                ``"num_training_days"`` : `int`
                    Number of days for training.
                ``"start_year"`` : `int`
                    Start year of the training period.
                ``"end_year"`` : `int`
                    End year of the forecast period.
                ``"origin_for_time_vars"`` : `float`
                    Continuous time representation of the first date in ``df``.

            In this function,

                - ``start_year`` and ``end_year`` are used to define ``daily_event_df_dict``.
                - ``simple_freq`` and ``num_training_days`` are used to define ``fs_components_df``.
                - ``simple_freq`` and ``num_training_days`` are used to set default ``feature_sets_enabled``.
                - ``origin_for_time_vars`` is used to set default ``origin_for_time_vars``.
                - the other parameters are ignored

            It is okay if ``num_training_points``, ``num_training_days``, ``start_year``, ``end_year``
            are computed for a superset of ``df``. This allows CV splits and backtest, which train on
            partial data, to use the same data-aware model parameters as the forecast on all training data.

            If None, the values are computed for ``df``. This corresponds to using the same
            modeling *approach* on the CV splits and backtest from `forecast_pipeline`, without
            requiring the same parameters. In this case, make sure ``forecast_horizon`` is at
            least as large as the test period for the split, to ensure all holidays are captured.
        freq : `str` or None, optional, default `None`
            Frequency of input data.
            Used to compute ``time_properties`` only if ``time_properties is None``.
            Frequency strings can have multiples, e.g. '5H'.
            See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
            for a list of frequency aliases.
            If None, inferred by `pandas.infer_freq`.
            Provide this parameter if ``df`` has missing timepoints.
        forecast_horizon : `int` or None, optional, default `None`
            Number of periods to forecast into the future. Must be > 0.
            Used to compute ``time_properties`` only if ``time_properties is None``.
            If None, default is determined by input data frequency.
            Used to determine forecast end date, to pull the appropriate holiday data.
            Should be at least as large as the prediction period (if this function
            is called from ``forecast_pipeline``, the prediction period for different
            splits is set via ``cv_horizon``, ``test_horizon``, ``forecast_horizon``).
        origin_for_time_vars : `float` or None, optional, default `None`
            The time origin used to create continuous variables for time.
            If None, uses the value from ``time_properties``.
        train_test_thresh : `datetime.datetime` or None, optional, default `None`
            e.g. datetime.datetime(2019, 6, 30)
            The threshold for training and testing split.
            Note that the final returned model is trained using all data.
            If None, training split is based on ``training_fraction``.
        training_fraction : `float` or None, optional, default 0.9
            The fraction of data used for training (0.0 to 1.0)
            Used only if ``train_test_thresh is None``.
            If this is also None or 1.0, then we skip testing
            and train on the entire dataset.
        fit_algorithm : `str`, optional, default "linear"
            The type of predictive model used in fitting.

            See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
            for available options and their parameters.
        fit_algorithm_params : `dict` or None, optional, default None
            Parameters passed to the requested fit_algorithm.
            If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
        holiday_lookup_countries : `list` [`str`] or "auto" or None, optional, default "auto"
            The countries that contain the holidays you intend to model
            (``holidays_to_model_separately``).

                * If "auto", uses a default list of countries
                  that contain the default ``holidays_to_model_separately``.
                  See `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.HOLIDAY_LOOKUP_COUNTRIES_AUTO`.
                * If a list, must be a list of country names.
                * If None or an empty list, no holidays are modeled.

        holidays_to_model_separately : `list` [`str`] or "auto" or `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES` or None, optional, default "auto"  # noqa: E501
            Which holidays to include in the model.
            The model creates a separate key, value for each item in ``holidays_to_model_separately``.
            The other holidays in the countries are grouped together as a single effect.

                * If "auto", uses a default list of important holidays.
                  See `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.HOLIDAYS_TO_MODEL_SEPARATELY_AUTO`.
                * If `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES`,
                  uses all available holidays in ``holiday_lookup_countries``. This can often
                  create a model that has too many parameters, and should typically be avoided.
                * If a list, must be a list of holiday names.
                * If None or an empty list, all holidays in ``holiday_lookup_countries`` are grouped together
                  as a single effect.

            Use ``holiday_lookup_countries`` to provide a list of countries where these holiday occur.
        holiday_pre_num_days : `int`, default 2
            Model holiday effects for ``holiday_pre_num_days`` days before the holiday.
        holiday_post_num_days : `int`, default 2
            Model holiday effects for ``holiday_post_num_days`` days after the holiday.
        holiday_pre_post_num_dict : `dict` [`str`, (`int`, `int`)] or None, default None
            Overrides ``pre_num`` and ``post_num`` for each holiday in
            ``holidays_to_model_separately``.
            For example, if ``holidays_to_model_separately`` contains "Thanksgiving" and "Labor Day",
            this parameter can be set to ``{"Thanksgiving": [1, 3], "Labor Day": [1, 2]}``,
            denoting that the "Thanksgiving" ``pre_num`` is 1 and ``post_num`` is 3, and "Labor Day"
            ``pre_num`` is 1 and ``post_num`` is 2.
            Holidays not specified use the default given by ``pre_num`` and ``post_num``.
        daily_event_df_dict : `dict` [`str`, `pandas.DataFrame`] or None, default None
            A dictionary of data frames, each representing events data for the corresponding key.
            Specifies additional events to include besides the holidays specified above. The format
            is the same as in `~greykite.algo.forecast.silverkite.SilverkiteForecast.forecast`.
            The DataFrame has two columns:

                - The first column contains event dates. Must be in a format
                  recognized by `pandas.to_datetime`. Must be at daily
                  frequency for proper join. It is joined against the time
                  in ``df``, converted to a day:
                  ``pd.to_datetime(pd.DatetimeIndex(df[time_col]).date)``.
                - the second column contains the event label for each date

            The column order is important; column names are ignored.
            The event dates must span their occurrences in both the training
            and future prediction period.

            During modeling, each key in the dictionary is mapped to a categorical variable
            named ``f"{EVENT_PREFIX}_{key}"``, whose value at each timestamp is specified
            by the corresponding DataFrame.

            For example, to manually specify a yearly event on September 1
            during a training/forecast period that spans 2020-2022::

                daily_event_df_dict = {
                    "custom_event": pd.DataFrame({
                        "date": ["2020-09-01", "2021-09-01", "2022-09-01"],
                        "label": ["is_event", "is_event", "is_event"]
                    })
                }

            It's possible to specify multiple events in the same df. Two events,
            ``"sep"`` and ``"oct"`` are specified below for 2020-2021::

                daily_event_df_dict = {
                    "custom_event": pd.DataFrame({
                        "date": ["2020-09-01", "2020-10-01", "2021-09-01", "2021-10-01"],
                        "event_name": ["sep", "oct", "sep", "oct"]
                    })
                }

            Use multiple keys if two events may fall on the same date. These events
            must be in separate DataFrames::

                daily_event_df_dict = {
                    "fixed_event": pd.DataFrame({
                        "date": ["2020-09-01", "2021-09-01", "2022-09-01"],
                        "event_name": "fixed_event"
                    }),
                    "moving_event": pd.DataFrame({
                        "date": ["2020-09-01", "2021-08-28", "2022-09-03"],
                        "event_name": "moving_event"
                    }),
                }

            The multiple event specification can be used even if events never overlap. An
            equivalent specification to the second example::

                daily_event_df_dict = {
                    "sep": pd.DataFrame({
                        "date": ["2020-09-01", "2021-09-01"],
                        "event_name": "is_event"
                    }),
                    "oct": pd.DataFrame({
                        "date": ["2020-10-01", "2021-10-01"],
                        "event_name": "is_event"
                    }),
                }

            Note: All these events are automatically added to the model. There is no need
            to specify them in ``extra_pred_cols`` as you would for
            `~greykite.algo.forecast.silverkite.SilverkiteForecast.forecast`.

            Note: Do not use `~greykite.common.constants.EVENT_DEFAULT`
            in the second column. This is reserved to indicate dates that do not
            correspond to an event.
        changepoints_dict : `dict` or None, optional, default None
            Specifies the changepoint configuration.

            ``"method"``: `str`
                The method to locate changepoints.
                Valid options:

                    - "uniform". Places n_changepoints evenly spaced changepoints to allow growth to change.
                    - "custom". Places changepoints at the specified dates.
                    - "auto". Automatically detects change points. For configuration, see
                      `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_trend_changepoints`

                Additional keys to provide parameters for each particular method are described below.
            ``"continuous_time_col"``: `str`, optional
                Column to apply ``growth_func`` to, to generate changepoint features
                Typically, this should match the growth term in the model
            ``"growth_func"``: callable or None, optional
                Growth function (scalar -> scalar). Changepoint features are created
                by applying ``growth_func`` to ``continuous_time_col`` with offsets.
                If None, uses identity function to use ``continuous_time_col`` directly
                as growth term
                If changepoints_dict["method"] == "uniform", this other key is required:

                    ``"n_changepoints"``: int
                        number of changepoints to evenly space across training period

                If changepoints_dict["method"] == "custom", this other key is required:

                    ``"dates"``: Iterable[Union[int, float, str, datetime]]
                        Changepoint dates. Must be parsable by pd.to_datetime.
                        Changepoints are set at the closest time on or after these dates
                        in the dataset.

                If changepoints_dict["method"] == "auto", the keys that matches the parameters in
                `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_trend_changepoints`,
                except ``df``, ``time_col`` and ``value_col``, are optional.
                Extra keys also include "dates", "combine_changepoint_min_distance" and "keep_detected" to specify
                additional custom trend changepoints. These three parameters correspond to the three parameters
                "custom_changepoint_dates", "min_distance" and "keep_detected" in
                `~greykite.algo.changepoint.adalasso.changepoints_utils.combine_detected_and_custom_trend_changepoints`.

        yearly_seasonality : `str` or `bool` or `int`
            Determines the yearly seasonality.
            'auto', True, False, or a number for the Fourier order
        quarterly_seasonality : `str` or `bool` or `int`
            Determines the quarterly seasonality.
            'auto', True, False, or a number for the Fourier order
        monthly_seasonality : `str` or `bool` or `int`
            Determines the monthly seasonality.
            'auto', True, False, or a number for the Fourier order
        weekly_seasonality : `str` or `bool` or `int`
            Determines the weekly seasonality.
            'auto', True, False, or a number for the Fourier order
        daily_seasonality : `str` or `bool` or `int`
            Determines the daily seasonality.
            'auto', True, False, or a number for the Fourier order
        max_daily_seas_interaction_order : `int` or None, optional, default `None`
            Max fourier order for interaction terms with daily seasonality.
            If None, uses all available terms.
        max_weekly_seas_interaction_order : `int` or None, optional, default `None`
            Max fourier order for interaction terms with weekly seasonality.
            If None, uses all available terms.
        autoreg_dict : `dict` or None, optional, default `None`
            A dictionary with arguments for `~greykite.common.features.timeseries_lags.build_autoreg_df`.
            That function's parameter ``value_col`` is inferred from the input of
            current function ``SilverkiteForecast::forecast``. Other keys are:

                ``"lag_dict"`` : `dict` or None
                ``"agg_lag_dict"`` : `dict` or None
                ``"series_na_fill_func"`` : callable

            See more details for above parameters in
            `~greykite.common.features.timeseries_lags.build_autoreg_df`.
        seasonality_changepoints_dict : `dict` or None, optional, default `None`
            The parameter dictionary for seasonality change point detection. Parameters are in
            `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`.
            Note ``df``, ``time_col``, ``value_col`` and ``trend_changepoints`` are auto populated,
            and do not need to be provided.
        min_admissible_value : `float` or None, optional, default `None`
            The minimum admissible value to return during prediction.
            If None, no limit is applied.
        max_admissible_value : `float` or None, optional, default `None`
            The maximum admissible value to return during prediction.
            If None, no limit is applied.
        uncertainty_dict : `dict` or None, optional, default `None`
            How to fit the uncertainty model. A dictionary with keys:
                ``"uncertainty_method"`` : `str`
                    The title of the method.
                    Only "simple_conditional_residuals" is implemented
                    in ``fit_prediction_model`` which calculates CIs using residuals
                ``"params"``: `dict`
                    A dictionary of parameters needed for
                    the requested ``uncertainty_method``. For example, for
                    ``uncertainty_method="simple_conditional_residuals"``, see
                    parameters of `~greykite.algo.uncertainty.conditional.conf_interval.conf_interval`,
                    listed briefly here:

                        ``"conditional_cols"``
                        ``"quantiles"``
                        ``"quantile_estimation_method"``
                        ``"sample_size_thresh"``
                        ``"small_sample_size_method"``
                        ``"small_sample_size_quantile"``

            If None, no uncertainty intervals are calculated.
        growth_term : `str` or None, optional, default "ct1"
            How to model the growth. Valid options are
            {"linear", "quadratic", "sqrt", "cuberoot"}.
        regressor_cols : `list` [`str`] or None, optional, default None
            The columns in ``df`` to use as regressors.
            These must be provided during prediction as well.
        feature_sets_enabled: `dict` [`str`, `bool` or "auto" or None] or `bool` or "auto" or None, default "auto"
            Whether to include interaction terms and categorical variables to increase model flexibility.

            If a `dict`, boolean values indicate whether include various sets of features in the model.
            The following keys are recognized
            (from `~greykite.algo.forecast.silverkite.constants.silverkite_column.SilverkiteColumn`):

                ``"COLS_HOUR_OF_WEEK"`` : `str`
                    Constant hour of week effect
                ``"COLS_WEEKEND_SEAS"`` : `str`
                    Daily seasonality interaction with is_weekend
                ``"COLS_DAY_OF_WEEK_SEAS"`` : `str`
                    Daily seasonality interaction with day of week
                ``"COLS_TREND_DAILY_SEAS"`` : `str`
                    Allow daily seasonality to change over time by is_weekend
                ``"COLS_EVENT_SEAS"`` : `str`
                    Allow sub-daily event effects
                ``"COLS_EVENT_WEEKEND_SEAS"`` : `str`
                    Allow sub-daily event effect to interact with is_weekend
                ``"COLS_DAY_OF_WEEK"`` : `str`
                    Constant day of week effect
                ``"COLS_TREND_WEEKEND"`` : `str`
                    Allow trend (growth, changepoints) to interact with is_weekend
                ``"COLS_TREND_DAY_OF_WEEK"`` : `str`
                    Allow trend to interact with day of week
                ``"COLS_TREND_WEEKLY_SEAS"`` : `str`
                    Allow weekly seasonality to change over time

            The following dictionary values are recognized:

                - True: include the feature set in the model
                - False: do not include the feature set in the model
                - None: do not include the feature set in the model
                - "auto" or not provided: use the default setting based on data frequency and size

            If not a `dict`:

                - if a boolean, equivalent to a dictionary with all values set to the boolean.
                - if None, equivalent to a dictionary with all values set to False.
                - if "auto", equivalent to a dictionary with all values set to "auto".

        extra_pred_cols : `list` [`str`] or None, optional, default `None`
            Columns to include in ``extra_pred_cols`` for ``SilverkiteForecast::forecast``.
            Other columns are added to ``extra_pred_cols`` by the other
            parameters of this function (i.e. ``holidays_*``, ``growth_term``,
            ``regressors``, ``feature_sets_enabled``).
            If `None`, treated is the same as [].
        regression_weight_col : `str` or None, default None
            The column name for the weights to be used in weighted regression version
            of applicable machine-learning models.
        simulation_based : `bool`, default False
            Boolean to specify if the future predictions are to be using simulations
            or not.
            Note that this is only used in deciding what parameters should be
            used for certain components e.g. autoregression, if automatic methods
            are requested. However, the auto-settings and the prediction settings
            regarding using simulations should match.

        Returns
        -------
        parameters : `dict`
            Parameters to call :func:`~greykite.algo.forecast.silverkite.SilverkiteForecast.forecast`.
        """
        if regressor_cols is None:
            regressor_cols = []

        if extra_pred_cols is None:
            extra_pred_cols = []
        else:
            # Does not modify the input list
            extra_pred_cols = extra_pred_cols.copy()

        if time_properties is None:
            # ``df`` only contains the dates for training,
            # so we can use ``use_univariate_ts=False``.
            # ``forecast_horizon`` must be at least as large as
            # the actual size of the test set / forecast set
            # in order to pull all holidays
            time_properties = get_forecast_time_properties(
                df=df,
                time_col=time_col,
                value_col=value_col,
                freq=freq,
                regressor_cols=regressor_cols,
                forecast_horizon=forecast_horizon)

        if time_properties is not None:
            forecast_horizon = forecast_horizon or time_properties.get("forecast_horizon")

        if origin_for_time_vars is None:
            origin_for_time_vars = time_properties["origin_for_time_vars"]

        # Specifies seasonality (added to ``extra_pred_cols`` by `SilverkiteForecast::forecast`)
        seasonality_dict = {
            "yearly_seasonality": yearly_seasonality,
            "quarterly_seasonality": quarterly_seasonality,
            "monthly_seasonality": monthly_seasonality,
            "weekly_seasonality": weekly_seasonality,
            "daily_seasonality": daily_seasonality,
        }

        fs_components_df = self.__get_silverkite_seasonality(
            simple_freq=time_properties["simple_freq"].name,
            num_days=time_properties["num_training_days"],
            seasonality=seasonality_dict)

        # Specifies growth (via ``extra_pred_cols``)
        growth_term_formula = None
        if growth_term is not None:
            growth_term_formula = GROWTH_COL_ALIAS[growth_term]
            extra_pred_cols += [growth_term_formula]

        # Specifies regressors (via ``extra_pred_cols``)
        if regressor_cols is not None:
            extra_pred_cols += regressor_cols

        # Specifies events (via ``daily_event_df_dict``, ``extra_pred_cols``).
        # Constant daily effect.
        holiday_df_dict = self.__get_silverkite_holidays(
            holiday_lookup_countries=holiday_lookup_countries,
            holidays_to_model_separately=holidays_to_model_separately,
            start_year=time_properties["start_year"],
            end_year=time_properties["end_year"],
            pre_num=holiday_pre_num_days,
            post_num=holiday_post_num_days,
            pre_post_num_dict=holiday_pre_post_num_dict)
        if holiday_df_dict is not None:
            # Adds holidays to the user-specified events,
            # giving preference to user events
            # if there are conflicts
            daily_event_df_dict = update_dictionary(
                holiday_df_dict,
                overwrite_dict=daily_event_df_dict)

        if not daily_event_df_dict:
            # Sets empty dictionary to None
            daily_event_df_dict = None

        extra_pred_cols += get_event_pred_cols(daily_event_df_dict)

        # Specifies ``extra_pred_cols`` (interactions and additional model terms).
        # Seasonality interaction order is limited by the available order and max requested.
        daily_seas_interaction_order = self.__get_seasonality_order_from_dataframe(
            seasonality=self._silverkite_seasonality_enum.DAILY_SEASONALITY.value,
            fs=fs_components_df,
            max_order=max_daily_seas_interaction_order
        )

        weekly_seas_interaction_order = self.__get_seasonality_order_from_dataframe(
            seasonality=self._silverkite_seasonality_enum.WEEKLY_SEASONALITY.value,
            fs=fs_components_df,
            max_order=max_weekly_seas_interaction_order
        )

        # updates `changepoints_dict`, unchanged if not "method" == "auto"
        changepoints_dict, changepoint_detector = get_changepoints_dict(
            df=df,
            time_col=time_col,
            value_col=value_col,
            changepoints_dict=changepoints_dict)

        # determines changepoint column names
        if changepoints_dict is not None:
            changepoints = get_changepoint_features_and_values_from_config(
                df=df,  # the training dataset
                time_col=time_col,
                changepoints_dict=changepoints_dict,
                origin_for_time_vars=origin_for_time_vars)
            changepoint_cols = changepoints["changepoint_cols"]
        else:
            changepoint_cols = []

        feature_sets_enabled = self.__get_feature_sets_enabled(
            simple_freq=time_properties["simple_freq"].name,
            num_days=time_properties["num_training_days"],
            feature_sets_enabled=feature_sets_enabled)

        model_feature_terms = self.__get_feature_sets_terms(
            daily_event_df_dict=daily_event_df_dict,
            daily_seas_interaction_order=daily_seas_interaction_order,
            weekly_seas_interaction_order=weekly_seas_interaction_order,
            growth_term=growth_term_formula,
            changepoint_cols=changepoint_cols)

        # extends ``extra_pred_cols`` by the requested feature sets from ``feature_sets_enabled``
        for feature_set_name, feature_set_terms in model_feature_terms.items():
            if feature_sets_enabled[feature_set_name]:
                extra_pred_cols += feature_set_terms
        extra_pred_cols = unique_elements_in_list(extra_pred_cols)

        # the parameters to call ``SilverkiteForecast::forecast``
        # parameters that are directly passed through are noted below
        parameters = dict(
            df=df,                                                          # pass-through
            time_col=time_col,                                              # pass-through
            value_col=value_col,                                            # pass-through
            origin_for_time_vars=origin_for_time_vars,
            extra_pred_cols=extra_pred_cols,
            train_test_thresh=train_test_thresh,                            # pass-through
            training_fraction=training_fraction,                            # pass-through
            fit_algorithm=fit_algorithm,                                    # pass-through
            fit_algorithm_params=fit_algorithm_params,                      # pass-through
            daily_event_df_dict=daily_event_df_dict,
            fs_components_df=fs_components_df,
            autoreg_dict=autoreg_dict,                                      # pass-through
            changepoints_dict=changepoints_dict,                            # pass-through
            seasonality_changepoints_dict=seasonality_changepoints_dict,    # pass-through
            changepoint_detector=changepoint_detector,
            min_admissible_value=min_admissible_value,                      # pass-through
            max_admissible_value=max_admissible_value,                      # pass-through
            uncertainty_dict=uncertainty_dict,                              # pass-through
            regression_weight_col=regression_weight_col,                    # pass-through
            forecast_horizon=forecast_horizon,                              # pass-through
            simulation_based=simulation_based                               # pass-through
        )

        return parameters

    def forecast_simple(
            self,
            *args,
            **kwargs):
        """A wrapper around ``SilverkiteForecast::forecast`` that simplifies some of the input parameters.

        Parameters
        ----------
        args : positional args
            Positional args to pass to
            :func:`~greykite.algo.forecast.silverkite.forecast_simple_silverkite.convert_simple_silverkite_params`.
            See that function for details.

        kwargs : keyword args
            Keyword args to pass to
            :func:`~greykite.algo.forecast.silverkite.forecast_simple_silverkite.convert_simple_silverkite_params`.
            See that function for details.

        Returns
        -------
        trained_model : `dict`
            The return value of :func:`~greykite.algo.forecast.silverkite.SilverkiteForecast.forecast`
            A dictionary that includes the fitted model from the function
            :func:`~greykite.algo.common.ml_models.fit_ml_model_with_evaluation`.
        """
        parameters = self.convert_params(*args, **kwargs)
        trained_model = super().forecast(**parameters)
        return trained_model

    def __get_requested_seasonality_order(
            self,
            requested_seasonality="auto",
            default_order=5,
            is_enabled_auto=True):
        """Returns requested seasonality fourier series order.

        Parameters
        ----------
        requested_seasonality :  `str` or `bool` or `int`, default = 'auto'
            The requested seasonality.
            'auto', True, False, or a number for the Fourier order.
        default_order : `int`
            The default order to use for 'auto' and True.
        is_enabled_auto : `bool`
            Whether the seasonality should be modeled for 'auto' seasonality.

        Returns
        -------
        order : `int`
            Seasonality fourier series order.
        """
        if requested_seasonality is True or (requested_seasonality == 'auto' and is_enabled_auto):
            order = default_order
        elif requested_seasonality is False or (requested_seasonality == 'auto' and not is_enabled_auto):
            order = 0
        else:
            try:
                order = int(requested_seasonality)
            except ValueError as e:
                log_message(f"Requested seasonality order '{requested_seasonality}' must be one of:"
                            f" 'auto', True, False, integer", LoggingLevelEnum.ERROR)
                raise e
        return order

    def __get_silverkite_seasonality(
            self,
            simple_freq=SimpleTimeFrequencyEnum.DAY.name,
            num_days=1000,
            seasonality=None):
        """Generates `fs_components_df` parameter for `forecast_silverkite`
        for modeling seasonality.

        Parameters
        ----------
        simple_freq : `str`
            SimpleTimeFrequencyEnum member that best matches the input data frequency
            according to `get_simple_time_frequency_from_period`
        num_days : `int`
            Number of days of observations in the input data
        seasonality : `dict` or None
            Seasonality configuration dictionary, with the following optional keys.
            (keys are SilverkiteSeasonalityEnum members in lower case):

                - ``"yearly_seasonality"`` : `str` or `bool` or `int` or None, default = 'auto'
                    Determines the yearly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"quarterly_seasonality"`` : `str` or `bool` or `int` or None, default = 'auto'
                    Determines the quarterly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"monthly_seasonality"`` : `str` or `bool` or `int` or None, default = 'auto'
                    Determines the monthly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"weekly_seasonality"`` : `str` or `bool` or `int` or None, default = 'auto'
                    Determines the weekly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"daily_seasonality"`` : `str` or `bool` or `int` or None, default = 'auto'
                    Determines the daily seasonality
                    'auto', True, False, or a number for the Fourier order

            None is equivalent to 'auto'. If 'auto', seasonality components are based on input data
            (``num_days``, ``simple_freq``), according to
            `~greykite.algo.forecast.silverkite.constants.silverkite_seasonality.SilverkiteSeasonalityEnum`.
            and `~greykite.algo.forecast.silverkite.constants.silverkite_time_frequency.SilverkiteTimeFrequencyEnum`.

        Returns
        -------
        fs_components_df : `pandas.DataFrame`
            Contains fourier series specification. Columns:

                - "name"
                - "period"
                - "order"
                - "seas_names"
        """
        if seasonality is None:
            seasonality = {}

        # recognized seasonalities for silverkite
        silverkite_seasonalities = self._silverkite_seasonality_enum.__members__.copy()
        silverkite_seasonalities = {k.lower(): v for k, v in silverkite_seasonalities.items()}

        # valid seasonalities based on input data frequency
        freq_valid_seas_names = SimpleTimeFrequencyEnum[simple_freq].value.valid_seas
        freq_auto_seas_names = self._silverkite_time_frequency_enum[simple_freq].value.auto_fourier_seas

        for key in seasonality.keys():
            if key not in silverkite_seasonalities.keys():
                raise ValueError(f"{key} must be one of {silverkite_seasonalities.keys()}")

        seasonalities = []  # seasonalities to add to the model
        for seas in silverkite_seasonalities.values():
            # keys are SilverkiteSeasonalityEnum members in lower case
            seas_input = seasonality.get(seas.name.lower(), "auto")
            # under auto configuration, seasonality is added if it's recommended for both
            # the input frequency and data size
            is_enabled_auto = (
                    num_days >= seas.value.default_min_days
                    and seas.name in freq_auto_seas_names)
            order = self.__get_requested_seasonality_order(
                requested_seasonality=seas_input,
                default_order=seas.value.order,
                is_enabled_auto=is_enabled_auto)
            if order > 0:
                if seas.name not in freq_valid_seas_names:
                    log_message(f"'{seas.name.lower()}' is typically not valid for "
                                f"data with '{simple_freq}' frequency. Each seasonality period "
                                f"should cover multiple observations in the data. To remove "
                                f"these seasonality terms from the model, remove {seas.name.lower()}={seas_input} "
                                f"or set it to 'auto' or 0.", LoggingLevelEnum.WARNING)
                seasonalities.append({
                    "name": seas.value.name,
                    "period": seas.value.period,
                    "order": order,  # user is allowed to override default order
                    "seas_names": seas.value.seas_names
                })

        # constructs dataframe where each seasonality is a row
        if len(seasonalities) > 0:
            fs = pd.DataFrame(
                seasonalities,
                columns=["name", "period", "order", "seas_names"])
        else:
            fs = None
        return fs

    def __get_seasonality_order_from_dataframe(
            self,
            seasonality,
            fs=None,
            max_order=None):
        """Returns fourier series order from a `pandas.DataFrame`
        fourier series specification. Return value is capped by ``max_order``.

        Parameters
        ----------
        seasonality : `SilverkiteSeasonalityEnum.Seasonality` namedtuple
            Which seasonality to extract from ``fs``.
            Has attributes ``name``, ``period``, ``order``, ``seas_names``
            Can be a `SilverkiteSeasonalityEnum` member value.
        fs : `pandas.DataFrame` or None, optional, default `None`
            Columns: "name", "period", "order", "seas_names"
            Suitable for ``fs_components_df`` parameter for ``forecast_silverkite``
            for modeling seasonality.
            Could be returned by ``get_silverkite_seasonality``.
            Assumes that ``name`` and ``seas_names`` uniquely identify a row.
        max_order: `int` or None, optional, default `None`
            Upper limit on seasonality_order.

        Returns
        -------
        fs_order : `int`
            The Fourier series order of the row with the given `name` and `seas_names`
        """
        order = 0
        if fs is not None:
            name_match = (fs["name"] == seasonality.name)
            seas_match = ((fs["seas_names"] == seasonality.seas_names)
                          if seasonality.seas_names is not None
                          else pd.isna(fs["seas_names"]))

            if any(name_match & seas_match):
                order = fs.loc[(name_match & seas_match), "order"].values[0]
        if max_order is not None:
            order = min(order, max_order)
        return order

    def __get_feature_sets_enabled(
            self,
            simple_freq=SimpleTimeFrequencyEnum.DAY.name,
            num_days=1000,
            feature_sets_enabled="auto"):
        """Returns default feature sets based on training data frequency and size.

        Parameters
        ----------
        simple_freq: `str`, default SimpleTimeFrequencyEnum.DAY.name
            SimpleTimeFrequencyEnum member that best matches the input data frequency
            according to `get_simple_time_frequency_from_period`
        num_days: `int`, default 1000
            Number of days of observations in the input data
        feature_sets_enabled: `dict` [`str`, `bool` or "auto" or None] or `bool` or "auto" or None, default "auto"
            Whether to include interaction terms and categorical variables to increase model flexibility.

            If a `dict`, boolean values indicate whether include various sets of features in the model.
            The following keys are recognized
            (from `~greykite.algo.forecast.silverkite.constants.silverkite_column.SilverkiteColumn`):

                ``"COLS_HOUR_OF_WEEK"`` : `str`
                    Constant hour of week effect
                ``"COLS_WEEKEND_SEAS"`` : `str`
                    Daily seasonality interaction with is_weekend
                ``"COLS_DAY_OF_WEEK_SEAS"`` : `str`
                    Daily seasonality interaction with day of week
                ``"COLS_TREND_DAILY_SEAS"`` : `str`
                    Allow daily seasonality to change over time by is_weekend
                ``"COLS_EVENT_SEAS"`` : `str`
                    Allow sub-daily event effects
                ``"COLS_EVENT_WEEKEND_SEAS"`` : `str`
                    Allow sub-daily event effect to interact with is_weekend
                ``"COLS_DAY_OF_WEEK"`` : `str`
                    Constant day of week effect
                ``"COLS_TREND_WEEKEND"`` : `str`
                    Allow trend (growth, changepoints) to interact with is_weekend
                ``"COLS_TREND_DAY_OF_WEEK"`` : `str`
                    Allow trend to interact with day of week
                ``"COLS_TREND_WEEKLY_SEAS"`` : `str`
                    Allow weekly seasonality to change over time

            The following dictionary values are recognized:

                - True: include the feature set in the model
                - False: do not include the feature set in the model
                - None: do not include the feature set in the model
                - "auto" or not provided: use the default setting based on data frequency and size

            If not a `dict`:

                - if a boolean, equivalent to a dictionary with all values set to the boolean.
                - if None, equivalent to a dictionary with all values set to False.
                - if "auto", equivalent to a dictionary with all values set to "auto".

        Returns
        -------
        feature_sets_enabled : `dict` [`str`, `bool`]
            Indicates which feature sets will be added to the model. Feature sets are determined
            by `get_model_feature_terms` and may be empty (e.g. if there are no events,
            there is no event:seasonality interaction)
            Same valid options as `feature_sets_enabled` parameter.
        """
        feature_sets_enabled_default = {
            self._silverkite_column.COLS_HOUR_OF_WEEK: False,
            self._silverkite_column.COLS_WEEKEND_SEAS: False,
            self._silverkite_column.COLS_DAY_OF_WEEK_SEAS: False,
            self._silverkite_column.COLS_TREND_DAILY_SEAS: False,
            self._silverkite_column.COLS_EVENT_SEAS: False,
            self._silverkite_column.COLS_EVENT_WEEKEND_SEAS: False,
            self._silverkite_column.COLS_DAY_OF_WEEK: False,
            self._silverkite_column.COLS_TREND_WEEKEND: False,
            self._silverkite_column.COLS_TREND_DAY_OF_WEEK: False,
            self._silverkite_column.COLS_TREND_WEEKLY_SEAS: False,
        }
        frequency = SimpleTimeFrequencyEnum[simple_freq].value

        # for sub-daily data
        if (
                frequency.seconds_per_observation
                <= SimpleTimeFrequencyEnum.HOUR.value.seconds_per_observation):
            if num_days >= TimeEnum.ONE_MONTH_IN_DAYS.value:
                # hour of week offset, helps the fourier terms
                feature_sets_enabled_default[self._silverkite_column.COLS_HOUR_OF_WEEK] = True
                # daily seasonality on weekday vs weekend
                feature_sets_enabled_default[self._silverkite_column.COLS_WEEKEND_SEAS] = True
                # daily seasonality by day of week
                feature_sets_enabled_default[self._silverkite_column.COLS_DAY_OF_WEEK_SEAS] = True
                # daily seasonality trend on weekday, weekend
                feature_sets_enabled_default[self._silverkite_column.COLS_TREND_DAILY_SEAS] = True

            if num_days < 3 * TimeEnum.ONE_YEAR_IN_DAYS.value:
                # holiday daily seasonality
                feature_sets_enabled_default[self._silverkite_column.COLS_EVENT_SEAS] = True
            else:
                # holiday daily seasonality that depends on weekend/weekday
                #   By pigeonhole principle, with reasonable assumption that a holiday must fall on a different
                #   day of the week for any three consecutive years (or else always be on the same day of week),
                #   it takes at most 3 years of training data to observe all weekend/weekday possibilities.
                feature_sets_enabled_default[self._silverkite_column.COLS_EVENT_WEEKEND_SEAS] = True

        # for sub-weekly data
        if (
                frequency.seconds_per_observation
                <= SimpleTimeFrequencyEnum.DAY.value.seconds_per_observation):
            # day of week offset, helps the fourier terms
            feature_sets_enabled_default[self._silverkite_column.COLS_DAY_OF_WEEK] = True

            # allows different trend on weekday vs weekend
            if num_days >= TimeEnum.ONE_MONTH_IN_DAYS.value:
                feature_sets_enabled_default[self._silverkite_column.COLS_TREND_WEEKEND] = True

            # allows trend interaction with day of week
            if num_days >= TimeEnum.ONE_QUARTER_IN_DAYS.value:
                feature_sets_enabled_default[self._silverkite_column.COLS_TREND_DAY_OF_WEEK] = True

            if num_days >= TimeEnum.ONE_YEAR_IN_DAYS.value:
                # weekly seasonality trend over time
                feature_sets_enabled_default[self._silverkite_column.COLS_TREND_WEEKLY_SEAS] = True

        # None is treated the same as False.
        # Intuitively, feature_sets_enabled=None should
        #   mean no feature sets are enabled.
        if feature_sets_enabled is None:
            feature_sets_enabled = False

        # Overrides defaults with user provided dictionary
        if feature_sets_enabled == "auto":
            pass  # uses the automatic defaults directly
        elif isinstance(feature_sets_enabled, bool):
            # All values are set to the provided boolean value
            for k in feature_sets_enabled_default.keys():
                feature_sets_enabled_default[k] = feature_sets_enabled
        elif isinstance(feature_sets_enabled, dict):
            # Uses the boolean values in `feature_sets_enabled` to override `feature_sets_enabled_default`
            for setting, is_enabled in feature_sets_enabled.items():
                if setting not in feature_sets_enabled_default:
                    raise ValueError(f"Unrecognized feature set: '{setting}'. Value feature sets are "
                                     f"{list(feature_sets_enabled_default.keys())}")

                if is_enabled == "auto":
                    # "auto" values are considered not set by the user and fall back to the default
                    continue
                if is_enabled is True:
                    # User explicitly turned on this feature set.
                    feature_sets_enabled_default[setting] = True
                elif is_enabled is False or is_enabled is None:
                    # User explicitly turned off this feature set.
                    # None values are treated the same as False.
                    feature_sets_enabled_default[setting] = False
                else:
                    raise ValueError(
                        f"Unrecognized `feature_sets_enabled` dictionary value for key {setting}: "
                        f"expected bool or 'auto' or None. Found: {is_enabled}")
        else:
            raise ValueError(
                f"Unrecognized type for `feature_sets_enabled`: expected bool, dict, 'auto', or None. Found: {feature_sets_enabled}")
        return feature_sets_enabled_default

    def __get_feature_sets_terms(
            self,
            daily_event_df_dict=None,
            daily_seas_interaction_order=0,
            weekly_seas_interaction_order=0,
            growth_term=None,
            changepoint_cols=None):
        """Defines features sets for use in the `extra_pred_cols` parameter
        to `forecast_silverkite`.
        Derived from events, seasonality, and trend (growth + changepoints).

        :param daily_event_df_dict: Optional[Dict[str, pd.DataFrame("date", "event")]]
            suitable for use as `daily_event_df_dict` parameter in `forecast_silverkite`
            Each event is modeled as its own effect
        :param daily_seas_interaction_order: int
            Order on interaction terms with daily seasonality
        :param weekly_seas_interaction_order: int
            Order on interaction terms with weekly seasonality
        :param growth_term: Optional[str]
            How to model the growth. Valid options are "linear", "quadratic", "sqrt", "cubic", "cuberoot"
        :param changepoint_cols: Optional[List[str]]
            Names of the changepoint feature columns to be generated by `build_silverkite_features`
        :return: Dict[str, List[str]]
            The patsy model terms for each feature set
            key: feature set name
            value: list of patsy model terms
                If there are no valid patsy model terms according to the input configuration,
                the list is empty.
                For example, if there are no events, the event related effects will be empty
        """
        # enumerates all possible keys
        extra_pred_cols_grouped = {
            self._silverkite_column.COLS_HOUR_OF_WEEK: [],
            self._silverkite_column.COLS_WEEKEND_SEAS: [],
            self._silverkite_column.COLS_DAY_OF_WEEK_SEAS: [],
            self._silverkite_column.COLS_TREND_DAILY_SEAS: [],
            self._silverkite_column.COLS_EVENT_SEAS: [],
            self._silverkite_column.COLS_EVENT_WEEKEND_SEAS: [],
            self._silverkite_column.COLS_DAY_OF_WEEK: [],
            self._silverkite_column.COLS_TREND_WEEKEND: [],
            self._silverkite_column.COLS_TREND_DAY_OF_WEEK: [],
            self._silverkite_column.COLS_TREND_WEEKLY_SEAS: [],
        }

        # the columns which constitute the trend
        if changepoint_cols is None:
            changepoint_cols = []
        growth_col = [growth_term] if growth_term is not None else []
        trend_cols = growth_col + changepoint_cols

        # all possible values of `dow` and `dow_hr` from `build_time_features_df`
        dow_levels = ["1-Mon", "2-Tue", "3-Wed", "4-Thu", "5-Fri", "6-Sat", "7-Sun"]
        dow_hr_levels = [f"{day + 1}_{str(hour).zfill(2)}" for day in range(7) for hour in range(24)]
        day_of_week = patsy_categorical_term(term="str_dow", levels=dow_levels)
        hour_of_week = patsy_categorical_term(term="dow_hr", levels=dow_hr_levels)

        extra_pred_cols_grouped[self._silverkite_column.COLS_DAY_OF_WEEK] = [day_of_week]
        extra_pred_cols_grouped[self._silverkite_column.COLS_HOUR_OF_WEEK] = [hour_of_week]
        extra_pred_cols_grouped[self._silverkite_column.COLS_TREND_WEEKEND] = [f"is_weekend:{col}" for col in trend_cols]
        extra_pred_cols_grouped[self._silverkite_column.COLS_TREND_DAY_OF_WEEK] = [f"{day_of_week}:{col}" for col in trend_cols]

        # allows major holidays to have different daily seasonality
        # interact with fourier series terms up to fs_daily_interaction_order
        daily_seasonality = self._silverkite_seasonality_enum.DAILY_SEASONALITY.value
        weekly_seasonality = self._silverkite_seasonality_enum.WEEKLY_SEASONALITY.value
        if daily_seas_interaction_order > 0:
            for holiday in self._silverkite_holiday.HOLIDAYS_TO_INTERACT:
                if daily_event_df_dict is not None and holiday in daily_event_df_dict.keys():
                    event_levels = [cst.EVENT_DEFAULT]  # reference level for non-event days, added by `add_daily_events`
                    # This event's levels
                    event_levels += list(daily_event_df_dict[holiday][cst.EVENT_DF_LABEL_COL].unique())

                    # `term` matches new_col in `add_daily_events`
                    term = f"{cst.EVENT_PREFIX}_{holiday}"
                    extra_pred_cols_grouped[self._silverkite_column.COLS_EVENT_SEAS] += cols_interact(
                        static_col=f"{patsy_categorical_term(term=term, levels=event_levels)}",
                        fs_name=daily_seasonality.name,
                        fs_order=daily_seas_interaction_order,
                        fs_seas_name=daily_seasonality.seas_names)

                    extra_pred_cols_grouped[self._silverkite_column.COLS_EVENT_WEEKEND_SEAS] += cols_interact(
                        static_col=f"is_weekend:{patsy_categorical_term(term=term, levels=event_levels)}",
                        fs_name=daily_seasonality.name,
                        fs_order=daily_seas_interaction_order,
                        fs_seas_name=daily_seasonality.seas_names)

            extra_pred_cols_grouped[self._silverkite_column.COLS_WEEKEND_SEAS] = cols_interact(
                static_col="is_weekend",
                fs_name=daily_seasonality.name,
                fs_order=daily_seas_interaction_order,
                fs_seas_name=daily_seasonality.seas_names)

            extra_pred_cols_grouped[self._silverkite_column.COLS_DAY_OF_WEEK_SEAS] = cols_interact(
                static_col=day_of_week,
                fs_name=daily_seasonality.name,
                fs_order=daily_seas_interaction_order,
                fs_seas_name=daily_seasonality.seas_names)

            for col in trend_cols:
                extra_pred_cols_grouped[self._silverkite_column.COLS_TREND_DAILY_SEAS] += cols_interact(
                    static_col=f"is_weekend:{col}",
                    fs_name=daily_seasonality.name,
                    fs_order=daily_seas_interaction_order,
                    fs_seas_name=daily_seasonality.seas_names)

        if weekly_seas_interaction_order > 0:
            for col in trend_cols:
                extra_pred_cols_grouped[self._silverkite_column.COLS_TREND_WEEKLY_SEAS] += cols_interact(
                    static_col=col,
                    fs_name=weekly_seasonality.name,
                    fs_order=weekly_seas_interaction_order,
                    fs_seas_name=weekly_seasonality.seas_names)

        return extra_pred_cols_grouped

    def __get_silverkite_holidays(
            self,
            holiday_lookup_countries="auto",
            holidays_to_model_separately="auto",
            start_year=2015,
            end_year=2030,
            pre_num=2,
            post_num=2,
            pre_post_num_dict=None):
        """Generates holidays dictionary for input to daily_event_df_dict parameter of silverkite model.
        The main purpose is to provide reasonable defaults for the holiday names and countries

        Parameters
        ----------
        holiday_lookup_countries : `list` [`str`] or "auto" or None, optional, default "auto"
            The countries that contain the holidays you intend to model
            (``holidays_to_model_separately``).

            * If "auto", uses a default list of countries
              that contain the default ``holidays_to_model_separately``.
              See `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.HOLIDAY_LOOKUP_COUNTRIES_AUTO`.
            * If a list, must be a list of country names.
            * If None or an empty list, no holidays are modeled.

        holidays_to_model_separately : `list` [`str`] or "auto" or `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES` or None, optional, default "auto"  # noqa: E501
            Which holidays to include in the model.
            The model creates a separate key, value for each item in ``holidays_to_model_separately``.
            The other holidays in the countries are grouped together as a single effect.

            * If "auto", uses a default list of important holidays.
              See `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.HOLIDAYS_TO_MODEL_SEPARATELY_AUTO`.
            * If `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES`,
              uses all available holidays in ``holiday_lookup_countries``. This can often
              create a model that has too many parameters, and should typically be avoided.
            * If a list, must be a list of holiday names.
            * If None or an empty list, all holidays in ``holiday_lookup_countries`` are grouped together
              as a single effect.

            Use ``holiday_lookup_countries`` to provide a list of countries where these holiday occur.
        start_year : `int`
            Year of first training data point, used to generate holiday events.
        end_year : `int`
            Year of last forecast data point, used to generate holiday events.
        pre_num : `int`
            Model holiday effects for ``pre_num`` days before the holiday.
        post_num : `int`
            Model holiday effects for ``post_num`` days after the holiday.
        pre_post_num_dict : `dict` [`str`, (`int`, `int`)] or None, default None
            Overrides ``pre_num`` and ``post_num`` for each holiday in
            ``holidays_to_model_separately``.
            For example, if ``holidays_to_model_separately`` contains "Thanksgiving" and "Labor Day",
            this parameter can be set to ``{"Thanksgiving": [1, 3], "Labor Day": [1, 2]}``,
            denoting that the "Thanksgiving" ``pre_num`` is 1 and ``post_num`` is 3, and "Labor Day"
            ``pre_num`` is 1 and ``post_num`` is 2.
            Holidays not specified use the default given by ``pre_num`` and ``post_num``.

        Returns
        -------
        daily_event_df_dict : `dict` [`str`, `pandas.DataFrame` [EVENT_DF_DATE_COL, EVENT_DF_LABEL_COL]]
            Suitable for use as `daily_event_df_dict` parameter in `forecast_silverkite`.
            Each holiday is modeled as its own effect (not specific to each country).

        See Also
        --------
        `~greykite.common.features.timeseries_features.get_available_holiday_lookup_countries`
        to list available countries for modeling.

        `~greykite.common.features.timeseries_features.get_available_holidays_across_countries`
        to see available holidays in those countries.
        """
        if holiday_lookup_countries is None:
            # `None` will not model any holidays
            holiday_lookup_countries = []
        elif holiday_lookup_countries == "auto":
            # countries that contain the default `holidays_to_model_separately`
            holiday_lookup_countries = self._silverkite_holiday.HOLIDAY_LOOKUP_COUNTRIES_AUTO
        elif not isinstance(holiday_lookup_countries, (list, tuple)):
            raise ValueError(
                f"`holiday_lookup_countries` should be a list, found {holiday_lookup_countries}")

        if holidays_to_model_separately is None:
            holidays_to_model_separately = []
        elif holidays_to_model_separately == "auto":
            # important holidays
            holidays_to_model_separately = self._silverkite_holiday.HOLIDAYS_TO_MODEL_SEPARATELY_AUTO
        elif holidays_to_model_separately == self._silverkite_holiday.ALL_HOLIDAYS_IN_COUNTRIES:
            holidays_to_model_separately = get_available_holidays_across_countries(
                countries=holiday_lookup_countries,
                year_start=start_year - 1,
                year_end=end_year + 1)
        elif not isinstance(holidays_to_model_separately, (list, tuple)):
            raise ValueError(
                f"`holidays_to_model_separately` should be a list, found {holidays_to_model_separately}")

        return generate_holiday_events(
            countries=holiday_lookup_countries,
            holidays_to_model_separately=holidays_to_model_separately,
            year_start=start_year - 1,  # subtract 1 just in case, to ensure coverage of all holidays
            year_end=end_year + 1,  # add 1 just in case, to ensure coverage of all holidays
            pre_num=pre_num,
            post_num=post_num,
            pre_post_num_dict=pre_post_num_dict)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""sklearn estimator for ``forecast_silverkite``"""

import pandas as pd
from sklearn.metrics import mean_squared_error

from greykite.algo.forecast.silverkite.forecast_silverkite import SilverkiteForecast
from greykite.algo.forecast.silverkite.silverkite_diagnostics import SilverkiteDiagnostics
from greykite.common import constants as cst
from greykite.common.python_utils import update_dictionary
from greykite.sklearn.estimator.base_silverkite_estimator import BaseSilverkiteEstimator


class SilverkiteEstimator(BaseSilverkiteEstimator):
    """Wrapper for `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

    Parameters
    ----------
    score_func : callable, optional, default mean_squared_error
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    coverage : `float` between [0.0, 1.0] or None, optional
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    null_model_params : `dict` or None, optional
        Dictionary with arguments to define ``DummyRegressor`` null model, default is `None`.
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    fit_algorithm_dict : `dict` or None, optional
        How to fit the model. A dictionary with the following optional keys.

            ``"fit_algorithm"`` : `str`, optional, default "linear"
                The type of predictive model used in fitting.

                See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
                for available options and their parameters.
            ``"fit_algorithm_params"`` : `dict` or None, optional, default None
                Parameters passed to the requested fit_algorithm.
                If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.

    uncertainty_dict : `dict` or `str` or None, optional
        How to fit the uncertainty model.
        See `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
        Note that this is allowed to be "auto". If None or "auto", will be set to
        a default value by ``coverage`` before calling ``forecast_silverkite``.
        See ``BaseForecastEstimator`` for details.
    fs_components_df : `pandas.DataFrame` or None, optional
        A dataframe with information about fourier series generation.
        If provided, it must contain columns with following names:

        - `"name"`: name of the timeseries feature (e.g. ``tod``, ``tow`` etc.).
        - `"period"`: Period of the fourier series.
        - `"order"`: Order of the fourier series.
          `"seas_names"`: Label for the type of seasonality (e.g. ``daily``, ``weekly`` etc.)
          and should be unique.
        - `~greykite.sklearn.estimator.silverkite_estimator.SilverkiteEstimator.validate_fs_components_df`
          checks for it, so that component plots don't have duplicate y-axis labels.

        This differs from the expected input of `forecast_silverkite` where `"period"`, `"order"`
        and `"seas_names"` are optional. This restriction is to facilitate appropriate computation
        of component (e.g. trend, seasonalities and holidays) effects. See Notes section in this
        docstring for a more detailed explanation with examples.


    Other parameters are the same as in
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

    If this Estimator is called from
    :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`,
    ``train_test_thresh`` and ``training_fraction`` should almost
    always be `None`, because train/test is handled outside
    this Estimator.

    The attributes are the same as
    `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator`.

    See Also
    --------
    `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator`
        For details on fit, predict, and component plots.
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
        Functions performing the fit and predict.
    """
    def __init__(
            self,
            silverkite: SilverkiteForecast = SilverkiteForecast(),
            silverkite_diagnostics: SilverkiteDiagnostics = SilverkiteDiagnostics(),
            score_func=mean_squared_error,
            coverage=None,
            null_model_params=None,
            origin_for_time_vars=None,
            extra_pred_cols=None,
            train_test_thresh=None,
            training_fraction=None,
            fit_algorithm_dict=None,
            daily_event_df_dict=None,
            fs_components_df=pd.DataFrame({
                "name": ["tod", "tow", "conti_year"],
                "period": [24.0, 7.0, 1.0],
                "order": [3, 3, 5],
                "seas_names": ["daily", "weekly", "yearly"]}),
            autoreg_dict=None,
            changepoints_dict=None,
            seasonality_changepoints_dict=None,
            changepoint_detector=None,
            min_admissible_value=None,
            max_admissible_value=None,
            uncertainty_dict=None,
            normalize_method=None,
            adjust_anomalous_dict=None,
            impute_dict=None,
            regression_weight_col=None,
            forecast_horizon=None,
            simulation_based=False):
        # every subclass of BaseSilverkiteEstimator must call super().__init__
        super().__init__(
            silverkite=silverkite,
            silverkite_diagnostics=silverkite_diagnostics,
            score_func=score_func,
            coverage=coverage,
            null_model_params=null_model_params,
            uncertainty_dict=uncertainty_dict)

        # necessary to set parameters, to ensure get_params() works
        # (used in grid search)
        self.score_func = score_func
        self.coverage = coverage
        self.null_model_params = null_model_params
        self.origin_for_time_vars = origin_for_time_vars
        self.extra_pred_cols = extra_pred_cols
        self.train_test_thresh = train_test_thresh
        self.fit_algorithm_dict = fit_algorithm_dict
        self.training_fraction = training_fraction
        self.daily_event_df_dict = daily_event_df_dict
        self.fs_components_df = fs_components_df
        self.autoreg_dict = autoreg_dict
        self.changepoints_dict = changepoints_dict
        self.seasonality_changepoints_dict = seasonality_changepoints_dict
        self.changepoint_detector = changepoint_detector
        self.min_admissible_value = min_admissible_value
        self.max_admissible_value = max_admissible_value
        self.uncertainty_dict = uncertainty_dict
        self.normalize_method = normalize_method
        self.adjust_anomalous_dict = adjust_anomalous_dict
        self.impute_dict = impute_dict
        self.regression_weight_col = regression_weight_col
        self.forecast_horizon = forecast_horizon
        self.simulation_based = simulation_based
        self.validate_inputs()

    def validate_inputs(self):
        """Validates the inputs to ``SilverkiteEstimator``."""
        # verifies that user-provided ``fs_components_df`` satisfies
        # the requirements of BaseSilverkiteEstimator
        if self.fs_components_df is not None:
            self.validate_fs_components_df(self.fs_components_df)

    def fit(
            self,
            X,
            y=None,
            time_col=cst.TIME_COL,
            value_col=cst.VALUE_COL,
            **fit_params):
        """Fits ``Silverkite`` forecast model.

        Parameters
        ----------
        X: `pandas.DataFrame`
            Input timeseries, with timestamp column,
            value column, and any additional regressors.
            The value column is the response, included in
            ``X`` to allow transformation by `sklearn.pipeline`.
        y: ignored
            The original timeseries values, ignored.
            (The ``y`` for fitting is included in ``X``).
        time_col: `str`
            Time column name in ``X``.
        value_col: `str`
            Value column name in ``X``.
        fit_params: `dict`
            additional parameters for null model.
        """
        # Initializes `fit_algorithm_dict` with default values.
        # This cannot be done in __init__ to remain compatible
        # with sklearn grid search.
        default_fit_algorithm_dict = {
            "fit_algorithm": "linear",
            "fit_algorithm_params": None}
        self.fit_algorithm_dict = update_dictionary(
            default_fit_algorithm_dict,
            overwrite_dict=self.fit_algorithm_dict)

        # fits null model
        super().fit(
            X=X,
            y=y,
            time_col=time_col,
            value_col=value_col,
            **fit_params)

        self.model_dict = self.silverkite.forecast(
            df=X,
            time_col=time_col,
            value_col=value_col,
            origin_for_time_vars=self.origin_for_time_vars,
            extra_pred_cols=self.extra_pred_cols,
            train_test_thresh=self.train_test_thresh,
            training_fraction=self.training_fraction,
            fit_algorithm=self.fit_algorithm_dict["fit_algorithm"],
            fit_algorithm_params=self.fit_algorithm_dict["fit_algorithm_params"],
            daily_event_df_dict=self.daily_event_df_dict,
            fs_components_df=self.fs_components_df,
            autoreg_dict=self.autoreg_dict,
            changepoints_dict=self.changepoints_dict,
            seasonality_changepoints_dict=self.seasonality_changepoints_dict,
            changepoint_detector=self.changepoint_detector,
            min_admissible_value=self.min_admissible_value,
            max_admissible_value=self.max_admissible_value,
            uncertainty_dict=self.uncertainty_dict,
            normalize_method=self.normalize_method,
            adjust_anomalous_dict=self.adjust_anomalous_dict,
            impute_dict=self.impute_dict,
            regression_weight_col=self.regression_weight_col,
            forecast_horizon=self.forecast_horizon,
            simulation_based=self.simulation_based)
        # sets attributes based on ``self.model_dict``
        super().finish_fit()

        return self

    @staticmethod
    def validate_fs_components_df(fs_components_df):
        """Validates the inputs of a fourier series components dataframe
        called by ``SilverkiteEstimator`` to validate the input fs_components_df.

        Parameters
        ----------
        fs_components_df : `pandas.DataFrame`
            A DataFrame with information about fourier series generation.
            Must contain columns with following names:

            - "name": name of the timeseries feature (e.g. "tod", "tow" etc.)
            - "period": Period of the fourier series
            - "order": Order of the fourier series
            - "seas_names": seas_name corresponding to the name (e.g. "daily", "weekly" etc.).

        """
        fs_cols_expected = ["name", "period", "order", "seas_names"]
        fs_cols_not_found = [col for col in fs_cols_expected if col not in fs_components_df.columns]
        if fs_cols_not_found:
            raise ValueError(f"fs_components_df is missing the following columns: {fs_cols_not_found}")

        if any(fs_components_df.duplicated(subset=["name", "seas_names"])):
            raise ValueError("Found multiple rows in fs_components_df with the same `names` and "
                             "`seas_names`. Make sure these are unique.")



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# original author: Rachit Arora

from dataclasses import dataclass
from enum import Enum
from typing import Type


@dataclass
class SilverkiteComponent:
    """Defines groupby time feature, xlabel and ylabel for Silverkite Component Plots."""
    groupby_time_feature: str

    xlabel: str

    ylabel: str


class SilverkiteComponentsEnum(Enum):
    """Defines groupby time feature, xlabel and ylabel for Silverkite Component Plots."""

    DAILY_SEASONALITY: SilverkiteComponent = SilverkiteComponent(
        groupby_time_feature="tod",
        xlabel="Hour of day",
        ylabel="daily")

    WEEKLY_SEASONALITY: SilverkiteComponent = SilverkiteComponent(
        groupby_time_feature="tow",
        xlabel="Day of week",
        ylabel="weekly")

    MONTHLY_SEASONALITY: SilverkiteComponent = SilverkiteComponent(
        groupby_time_feature="tom",
        xlabel="Time of month",
        ylabel="monthly")

    QUARTERLY_SEASONALITY: SilverkiteComponent = SilverkiteComponent(
        groupby_time_feature="toq",
        xlabel="Time of quarter",
        ylabel="quarterly")

    YEARLY_SEASONALITY: SilverkiteComponent = SilverkiteComponent(
        groupby_time_feature="toy",
        xlabel="Time of year",
        ylabel="yearly")


class SilverkiteComponentsEnumMixin:
    """Provides a mixin for the SilverkiteComponentsEnum constants"""

    def get_silverkite_components_enum(self) -> Type[SilverkiteComponentsEnum]:
        """Return the SilverkiteComponentsEnum constants"""
        return SilverkiteComponentsEnum



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# original author: Rachit Arora

from dataclasses import dataclass
from enum import Enum
from typing import Type

from greykite.common.enums import FrequencyEnum
from greykite.common.enums import SeasonalityEnum


@dataclass
class SilverkiteFrequency:
    """Provides properties for modeling for various time frequencies in Silverkite."""

    name: FrequencyEnum
    """The name of the FrequencyEnum e.g. MINUTE, HOUR etc."""

    auto_fourier_seas: [SeasonalityEnum]
    """The recommended seasonality periods to model with fourier series"""


class SilverkiteTimeFrequencyEnum(Enum):
    """Provides properties for modeling for various time frequencies in Silverkite.
    The enum names is the time frequency, corresponding to the simple time
    frequencies in `~greykite.common.enums.SimpleTimeFrequencyEnum`.
    """

    # minutely
    MINUTE: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.MINUTE,
        auto_fourier_seas={SeasonalityEnum.DAILY_SEASONALITY.name,
                           SeasonalityEnum.WEEKLY_SEASONALITY.name,
                           # MONTHLY_SEASONALITY is excluded from defaults
                           SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                           SeasonalityEnum.YEARLY_SEASONALITY.name})

    # hourly
    HOUR: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.HOUR,
        auto_fourier_seas={SeasonalityEnum.DAILY_SEASONALITY.name,
                           SeasonalityEnum.WEEKLY_SEASONALITY.name,
                           # MONTHLY_SEASONALITY is excluded from defaults
                           SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                           SeasonalityEnum.YEARLY_SEASONALITY.name})

    # daily
    DAY: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.DAY,
        auto_fourier_seas={SeasonalityEnum.WEEKLY_SEASONALITY.name,
                           # MONTHLY_SEASONALITY is excluded from defaults
                           SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                           SeasonalityEnum.YEARLY_SEASONALITY.name})

    # weekly
    WEEK: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.WEEK,
        auto_fourier_seas={SeasonalityEnum.MONTHLY_SEASONALITY.name,
                           SeasonalityEnum.QUARTERLY_SEASONALITY.name,
                           SeasonalityEnum.YEARLY_SEASONALITY.name})

    # monthly
    MONTH: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.MONTH,
        auto_fourier_seas={
            # QUARTERLY_SEASONALITY and YEARLY_SEASONALITY are excluded from defaults
            # It's better to use `C(month)` as a categorical feature indicating the month
        })

    # quarterly
    QUARTER: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.QUARTER,
        auto_fourier_seas={
            # YEARLY_SEASONALITY is excluded from defaults
            # It's better to use `C(quarter)` as a categorical feature indicating the quarter
        })

    # yearly
    YEAR: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.YEAR,
        auto_fourier_seas={})

    # longer than yearly
    MULTIYEAR: SilverkiteFrequency = SilverkiteFrequency(
        name=FrequencyEnum.MULTIYEAR,
        auto_fourier_seas={})


class SilverkiteTimeFrequencyEnumMixin:
    """Provides a mixin for the SilverkiteTimeFrequencyEnum constants"""

    def get_silverkite_time_frequency_enum(self) -> Type[SilverkiteTimeFrequencyEnum]:
        """Return the SilverkiteTimeFrequencyEnum constants"""
        return SilverkiteTimeFrequencyEnum



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Functions that extract timeseries properties, with additional
domain logic for forecasting.
"""

import math
from datetime import timedelta

from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.enums import SimpleTimeFrequencyEnum
from greykite.common.enums import TimeEnum
from greykite.common.features.timeseries_features import get_default_origin_for_time_vars
from greykite.common.time_properties import get_canonical_data
from greykite.common.time_properties import min_gap_in_seconds


def get_default_horizon_from_period(period, num_observations=None):
    """Returns default forecast horizon based on input data period and num_observations
    :param period: float
        Period of each observation (i.e. average time between observations, in seconds)
    :param num_observations: Optional[int]
        Number of observations for training
    :return: int
        default number of periods to forecast
    """
    default_from_period = get_simple_time_frequency_from_period(period).value.default_horizon
    if num_observations is not None:
        default_from_observations = num_observations // 2  # twice as much training data as forecast horizon
        return min(default_from_period, default_from_observations)  # horizon based on limiting factor
    else:
        return default_from_period


def get_simple_time_frequency_from_period(period):
    """Returns SimpleTimeFrequencyEnum based on input data period
    :param period: float
        Period of each observation (i.e. average time between observations, in seconds)
    :return: SimpleTimeFrequencyEnum
        SimpleTimeFrequencyEnum is used to define default values for horizon, seasonality, etc.
        (but original data frequency is not modified)
    """
    freq_threshold = [
        (SimpleTimeFrequencyEnum.MINUTE, 10.05),  # <= 10 minutes is considered minute-level, buffer for abnormalities
        (SimpleTimeFrequencyEnum.HOUR, 6.05),     # <= 6 hours is considered hourly, buffer for abnormalities
        (SimpleTimeFrequencyEnum.DAY, 2.05),      # <= 2 days is considered daily, buffer for daylight savings
        (SimpleTimeFrequencyEnum.WEEK, 2.05),     # <= 2 weeks is considered weekly, buffer for daylight savings
        (SimpleTimeFrequencyEnum.MONTH, 2.05),    # <= 2 months is considered monthly, buffer for 31-day month
        (SimpleTimeFrequencyEnum.YEAR, 1.01),     # <= 1 years is considered yearly, buffer for leap year
    ]

    for simple_freq, threshold in freq_threshold:
        if period <= simple_freq.value.seconds_per_observation * threshold:
            return simple_freq
    return SimpleTimeFrequencyEnum.MULTIYEAR


def get_forecast_time_properties(
        df,
        time_col=TIME_COL,
        value_col=VALUE_COL,
        freq=None,
        regressor_cols=None,
        train_end_date=None,
        forecast_horizon=None):
    """Returns the number of training points in `df`, the start year, and prediction end year

    Parameters
    ----------
    df : `pandas.DataFrame` with columns [``time_col``, ``value_col``]
        Univariate timeseries data to forecast
    time_col : `str`, default ``TIME_COL`` in constants.py
        Name of timestamp column in df
    value_col : `str`, default ``VALUE_COL`` in constants.py
        Name of value column in df (the values to forecast)
    freq : `str` or None, default None
        Frequency of input data. Used to generate future dates for prediction.
        Frequency strings can have multiples, e.g. '5H'.
        See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        for a list of frequency aliases.
        If None, inferred by pd.infer_freq.
        Provide this parameter if ``df`` has missing timepoints.
    regressor_cols : `list` [`str`] or None, optional, default None
        A list of regressor columns used in the training and prediction DataFrames.
        If None, no regressor columns are used.
        Regressor columns that are unavailable in ``df`` are dropped.
    train_end_date : `datetime.datetime`, optional, default None
        Last date to use for fitting the model. Forecasts are generated after this date.
        If None, it is set to the last date with a non-null value in
        ``value_col`` of ``df``.
    forecast_horizon : `int` or None, default None
        Number of periods to forecast into the future. Must be > 0
        If None, default is determined from input data frequency

    Returns
    -------
    time_properties : `dict` [`str`, `any`]
        Time properties dictionary with keys:

        ``"period"`` : `int`
            Period of each observation (i.e. minimum time between observations, in seconds).
        ``"simple_freq"`` : `SimpleTimeFrequencyEnum`
            ``SimpleTimeFrequencyEnum`` member corresponding to data frequency.
        ``"num_training_points"`` : `int`
            Number of observations for training.
        ``"num_training_days"`` : `int`
            Number of days for training.
        ``"days_per_observation"``: `float`
            The time frequency in day units.
        ``"forecast_horizon"``: `int`
            The number of time intervals for which forecast is needed.
        ``"forecast_horizon_in_timedelta"``: `datetime.timedelta`
            The forecast horizon length in timedelta units.
        ``"forecast_horizon_in_days"``: `float`
            The forecast horizon length in day units.
        ``"start_year"`` : `int`
            Start year of the training period.
        ``"end_year"`` : `int`
            End year of the forecast period.
        ``"origin_for_time_vars"`` : `float`
            Continuous time representation of the first date in ``df``.
    """
    if regressor_cols is None:
        regressor_cols = []

    # Defines ``fit_df``, the data available for fitting the model
    # and its time column (in `datetime.datetime` format)
    canonical_data_dict = get_canonical_data(
        df=df,
        time_col=time_col,
        value_col=value_col,
        freq=freq,
        train_end_date=train_end_date,
        regressor_cols=regressor_cols)
    fit_df = canonical_data_dict["fit_df"]

    # Calculates basic time properties
    train_start = fit_df[TIME_COL].min()
    start_year = int(train_start.strftime("%Y"))
    origin_for_time_vars = get_default_origin_for_time_vars(fit_df, TIME_COL)
    period = min_gap_in_seconds(df=fit_df, time_col=TIME_COL)
    simple_freq = get_simple_time_frequency_from_period(period)
    num_training_points = fit_df.shape[0]

    # Calculates number of (fractional) days in the training set
    time_delta = fit_df[TIME_COL].max() - train_start
    num_training_days = (
            time_delta.days
            + (time_delta.seconds + period) / TimeEnum.ONE_DAY_IN_SECONDS.value)

    # Calculates forecast horizon (as a number of periods)
    if forecast_horizon is None:
        # expected to be kept in sync with default value set in ``get_default_time_parameters``
        forecast_horizon = get_default_horizon_from_period(
            period=period,
            num_observations=num_training_points)

    days_per_observation = period / TimeEnum.ONE_DAY_IN_SECONDS.value
    forecast_horizon_in_days = forecast_horizon * days_per_observation
    forecast_horizon_in_timedelta = timedelta(days=forecast_horizon_in_days)

    # Calculates forecast end year
    train_end = fit_df[TIME_COL].max()
    days_to_forecast = math.ceil(forecast_horizon * days_per_observation)
    future_end = train_end + timedelta(days=days_to_forecast)
    end_year = int(future_end.strftime("%Y"))

    return {
        "period": period,
        "simple_freq": simple_freq,
        "num_training_points": num_training_points,
        "num_training_days": num_training_days,
        "days_per_observation": days_per_observation,
        "forecast_horizon": forecast_horizon,
        "forecast_horizon_in_timedelta": forecast_horizon_in_timedelta,
        "forecast_horizon_in_days": forecast_horizon_in_days,
        "start_year": start_year,
        "end_year": end_year,
        "origin_for_time_vars": origin_for_time_vars
    }
